import inspect
import time

from AnyQt.QtCore import QThread, pyqtSignal

# This function allow to put on separate threads of the GUI some executions function, to allow the GUI to still be responsive
# While output is processed. 

''''
# Start progress bar
self.progressBarInit()

# Connect and start thread : main function, progress, result and finish
# --> progress is used in the main function to track progress (with a callback)
# --> result is used to collect the result from main function
# --> finish is just an empty signal to indicate that the thread is finished
# self.thread = thread_management.Thread(embeddings.create_embeddings, self.data, self.model_path)
self.thread = thread_management.Thread(embeddings.create_embeddings, self.data, self.model)
self.thread.progress.connect(self.handle_progress)
self.thread.result.connect(self.handle_result)
self.thread.finish.connect(self.handle_finish)
self.thread.start()
####-----
def handle_progress(self, value):
    self.progressBarSet(value)
def handle_result(self, result):
    try:
        self.Outputs.data.send(result)
    except Exception as e:
        print("An error occurred when sending out_data:", e)
        self.Outputs.data.send(None)
        return
def handle_finish(self):
    print("Embeddings finished")
    self.progressBarFinished()
'''


class Thread(QThread):
    progress = pyqtSignal(object)
    result = pyqtSignal(object)
    finish = pyqtSignal()

    def __init__(self, func, *args, **kwargs):
        super().__init__()
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.setTerminationEnabled(True)
        self.stop = False

    def run(self):
        inputs = inspect.getfullargspec(self.func).args
        final_args = list(self.args)
        final_kwargs = dict(self.kwargs)

        # Ajout des entrées optionnelles si elles sont attendues par la fonction
        if "progress_callback" in inputs:
            final_kwargs["progress_callback"] = self.progress.emit
        if "argself" in inputs:
            final_kwargs["argself"] = self

        result = self.func(*final_args, **final_kwargs)
        self.result.emit(result)
        self.finish.emit()

    def safe_quit(self):
        self.stop = True
        self.wait()

    def berserk_quit(self):
        self.stop = True
        time.sleep(1)
        if not self.isRunning():
            return
        self.terminate()
        self.wait()



import ctypes
import os
import sys
import Orange.data
from Orange.widgets.widget import Input, Output, OWWidget
from Orange.data import Table

if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic
    from Orange.widgets.orangecontrib.AAIT.utils.unlink_table_domain import unlink_domain
else:
    from orangecontrib.AAIT.utils.import_uic import uic
    from orangecontrib.AAIT.utils.unlink_table_domain import unlink_domain

class EndLoopWidget(OWWidget):
    name = "End Loop"
    description = "Widget to end a loop based on a predefined condition."
    icon = "icons/endloop.png"
    category = "AAIT - ALGORITHM"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/endloop.png"

    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owendloop.ui")
    want_control_area = False
    priority = 1011

    class Inputs:
        in_data = Input("Data In", Orange.data.Table, multiple=True)
        in_pointer = Input("End of the Loop Do-While", str, auto_summary=False)

    class Outputs:
        out_data = Output("Data Out", Orange.data.Table)

    def __init__(self):
        super().__init__()
        self.setFixedWidth(470)
        self.setFixedHeight(300)
        uic.loadUi(self.gui, self)

        self.data = None
        self.out_data = None
        self.in_pointer = None

    @Inputs.in_data
    def set_data(self, data,id=None):
        self.error("")
        if data is None:
            self.data = None
            return
        if self.in_pointer is None:
            return

        self.data = data

        # cas ou on boucle temps que le nombre de ligne augmente
        if "True"==self.get_is_allow_to_change_line_number_from_start():
            if self.get_nb_line() == self.get_nb_line_from_start():
                # on s'arrete
                self.reinitialize_iter()
                self.Outputs.out_data.send(self.data)
                return
            # on boucle
            ctypes.cast(int(self.in_pointer), ctypes.py_object).value.set_data(self.data)
            return

        # cas ou on boucle sur le nombre de ligne en entrée du workflow
        if "True" == self.iter_of_line_number_from_start():
            if self.out_data is None:
                self.out_data = unlink_domain(self.data)
            else:
                self.out_data = Table.concatenate([self.out_data, unlink_domain(self.data)])

            if self.get_nb_line_from_start() == self.get_iter():
                # on s'arrete
                self.reinitialize_iter()
                self.Outputs.out_data.send(self.out_data)
                self.out_data = None
                return
            # on boucle
            ctypes.cast(int(self.in_pointer), ctypes.py_object).value.execute_iter_of_line_number()
            return


        # Check if the number of lines has changed
        if self.get_nb_line() != self.get_nb_line_from_start():
            self.error("Error! You can't change the number of lines")
            return

        #check domain
        column_names_input_temp, column_types_input=self.get_column_name_and_type_from_start()
        column_names_output_temp, column_types_ouput = self.get_column_name_and_type()

        column_names_input=[]
        column_names_output = []
        for idx,element in enumerate(column_names_input_temp):
            column_names_input.append(element+ " ("+column_types_input[idx]+")")

        for idx, element in enumerate(column_names_output_temp):
            column_names_output.append(element + " (" + column_types_ouput[idx] + ")")

        if column_names_input != column_names_output:
            # Find elements that are only in one of the lists
            diff1 = [x for x in column_names_input if x not in column_names_output]
            diff2 = [x for x in column_names_output if x not in column_names_input]

            # Detect swapped elements
            swapped = [(x, y) for x, y in zip(column_names_input, column_names_output) if
                       x != y and x in column_names_output and y in column_names_input]

            # Function to format a list as a sentence
            def list_to_sentence(lst):
                if not lst:
                    return "None"
                if len(lst) == 1:
                    return lst[0]
                return ", ".join(lst[:-1]) + " and " + lst[-1]

            # Build the output message
            sentence = ""
            if diff1:
                sentence += f"The elements that are only in the beginning of the loop table are: {list_to_sentence(diff1)}.\n"
            if diff2:
                sentence += f"The elements that are only in the end of the loop table are: {list_to_sentence(diff2)}.\n"

            if swapped:
                swapped_sentence = ", ".join([f"{x} <-> {y}" for x, y in swapped])
                sentence += f"Detected swapped elements: {swapped_sentence}."
            self.error(sentence)
            return

        # input of the loop== output of the loop
        old_table=ctypes.cast(int(self.in_pointer), ctypes.py_object).value.get_in_data()
        idem=True
        for idx,element in enumerate(old_table):
            if element!=self.data[idx]:
                idem=False
                break

        # check if loop doesn't change anithing
        if idem:
            self.error("Error! your loop didn't change the data table")
            return

        if any(x > 0 for x in self.data.get_column("iter")):
            ctypes.cast(int(self.in_pointer), ctypes.py_object).value.set_data(self.data)
            return
        self.reinitialize_iter()
        self.Outputs.out_data.send(self.data)

    @Inputs.in_pointer
    def set_pointer(self, pointer):
        self.in_pointer = int(pointer) if pointer else None

    def get_column_name_and_type_from_start(self):
        if self.in_pointer is not None:
            result = ctypes.cast(int(self.in_pointer), ctypes.py_object).value.get_column_name_and_type()
        return result

    def get_nb_line_from_start(self):
        result = 0
        if self.in_pointer is not None:
            result = ctypes.cast(int(self.in_pointer), ctypes.py_object).value.get_nb_line()
        return result

    def get_is_allow_to_change_line_number_from_start(self):
        result = "False"
        if self.in_pointer is not None:
            result = ctypes.cast(int(self.in_pointer), ctypes.py_object).value.is_allow_to_change_line_number()
        return result

    def iter_of_line_number_from_start(self):
        result = "False"
        if self.in_pointer is not None:
            result = ctypes.cast(int(self.in_pointer), ctypes.py_object).value.iter_of_line_number()
        return result

    def get_iter(self):
        if self.in_pointer is not None:
            result = ctypes.cast(int(self.in_pointer), ctypes.py_object).value.get_iter()
        return result


    def reinitialize_iter(self):
        if self.in_pointer is not None:
            ctypes.cast(int(self.in_pointer), ctypes.py_object).value.reinitialize_iter()



    def get_nb_line(self):
        # Return the number of lines to compare with another widget
        if self.data is None:
            return 0
        return len(self.data)

    def get_column_name_and_type(self):
        # Return the name and type of 'data_in' to verify if they are the same
        if self.data is None:
            return [[], []]
        column_names = []
        column_types = []
        for element in self.data.domain:
            column_names.append(str(element))
            column_types.append(str(type(element)))
        return column_names, column_types





if __name__ == "__main__":
    from AnyQt.QtWidgets import QApplication
    app = QApplication(sys.argv)
    obj = EndLoopWidget()
    obj.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()




import os

prompt_templates = {
    "llama":
"""<|begin_of_text|><|start_header_id|>system<|end_header_id|>

{system_prompt}<|eot_id|><|start_header_id|>user<|end_header_id|>

{user_prompt}<|eot_id|><|start_header_id|>assistant<|end_header_id|>""",



    "mistral": """<s>[INST] {user_prompt} [/INST]</s>\n""",



    "solar":
"""### User: {user_prompt}

### Assistant: {assistant_prompt}""",



    "deepseek":
"""### Instruction: {user_prompt}

### Response:""",



    "qwen":
"""<|im_start|>system
{system_prompt}<|im_end|>
<|im_start|>user
{user_prompt}<|im_end|>
<|im_start|>assistant
{assistant_prompt}""",




    "gemma":
"""<start_of_turn>user
{user_prompt}<end_of_turn>
<start_of_turn>model
{assistant_prompt}
""",
    
    
    
    "granite":
"""<|system|>
{system_prompt}
<|user|>
{user_prompt}
<|assistant|>
{assistant_prompt}
""",



    "phi":
"""<|im_start|>system<|im_sep|>
{system_prompt}<|im_end|>
<|im_start|>user<|im_sep|>
{user_prompt}<|im_end|>
<|im_start|>assistant<|im_sep|>
{assistant_prompt}
"""
}


stop_tokens = {
    "llama": "<|eot_id|>",
    "mistral": "</s>",
    "qwen": "<|im_end|>",
    "gemma": "<end_of_turn>",
    "granite": "<|endoftext|>",
    "phi": "<|im_end|>"
}


model_types = {
    "solar-10.7b-instruct-v1.0.Q6_K.gguf": "solar",
    "solar-10.7b-instruct-v1.0-uncensored.Q6_K.gguf": "solar",
    "Mistral-7B-Instruct-v0.3.Q6_K.gguf": "mistral",
    "Qwen2.5.1-Coder-7B-Instruct-Q6_K.gguf": "qwen",
    "qwen2.5-3b-instruct-q4_k_m.gguf": "qwen",
    "deepseek-coder-6.7b-instruct.Q6_K.gguf": "deepseek"
}

model_keywords = ["qwen", "solar", "mistral", "llama", "deepseek", "gemma", "granite", "phi"]

def apply_prompt_template(model_name, user_prompt, assistant_prompt="", system_prompt=""):
    """
    Apply a prompt template based on the given model name and user input.

    Parameters:
        model_name (str): The name of the model used to determine its type.
        user_prompt (str): The user input or request to embed into the prompt.
        assistant_prompt (str, optional): The assistant's beginning of response, if any. Defaults to an empty string.
        system_prompt (str, optional): A system-level instruction or context to include in the prompt. Defaults to an empty string.

    Returns:
        str: The formatted prompt that is ready to be passed to the model.
    """
    model_name = os.path.basename(model_name)
    model_type = model_types.get(model_name)
    if model_type is None:
        model_type = next((keyword for keyword in model_keywords if keyword in model_name.lower()))

    # Retrieve the template and inject the prompts
    template = prompt_templates.get(model_type, "{user_prompt}")  # Default template if none found
    # Format the template
    prompt = template.format(user_prompt=user_prompt, assistant_prompt=assistant_prompt, system_prompt=system_prompt)
    return prompt


def get_stop_token(model_name):
    """
    Get the stop token according to the model name / type.

    Parameters:
        model_name (str): The name of the model used to determine its type.
    """
    # If there is a stop token
    try:
        # Get the model type
        model_type = model_types[model_name]
        # Get the template for the model type
        stop_token = stop_tokens[model_type]
    except KeyError as e:
        print(f"Your model {model_name} has no stop token defined. See prompt_management.py. (detail: {e})")
        return None
    return stop_token








import os
import json
import ast
from Orange.widgets import  widget
from Orange.widgets.settings import Setting
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.widgets.widget import Input, Output
from Orange.data import Table
from AnyQt.QtWidgets import QCheckBox
import copy

if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic
else:
    from orangecontrib.AAIT.utils.import_uic import uic

class OWSaveFilepathEntry(widget.OWWidget):
    name = "Save with Filepath Entry"
    description = "Save data to a .pkl file, based on the provided path"
    category = "AAIT - TOOLBOX"
    icon = "icons/owsavefilepathentry.svg"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/owsavefilepathentry.svg"
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owsavewithpath.ui")
    priority = 1220
    want_control_area = False

    # Persistent settings for fileId and CSV delimiter
    filename: str = Setting("embeddings.pkl")
    annotations: bool = Setting(True)
    purge_widget: bool = Setting(False)


    class Inputs:
        data = Input("Data", Table)
        save_path = Input("Path", str, auto_summary=False)
        path_table = Input("Path Table", Table)

    class Outputs:
        data = Output("Data", Table)

    @Inputs.data
    def dataset(self, data): 
        """Handle new data input."""
        if data is None:
            self.data=None
            return
        self.data = data
        self.run()


    @Inputs.save_path
    def set_save_path(self, in_save_path):
        if in_save_path is None:
            self.save_path=None
            return
        self.save_path = in_save_path.replace('"', '')
        self.json = False
        if self.save_path.endswith(".json"):
            self.json = True
        self.run()

    @Inputs.path_table
    def set_path_table(self, in_path_table):
        if in_path_table is None:
            self.save_path=None
            return
        self.json = False
        if "path" in in_path_table.domain:
            if in_path_table[0]["path"].value.endswith(".json"):
                self.json = True
            self.save_path = in_path_table[0]["path"].value.replace('"', '')
            self.run()


    def __init__(self):
        super().__init__()
        # Qt Management
        self.setFixedWidth(470)
        self.setFixedHeight(300)
        uic.loadUi(self.gui, self)
        self.checkbox_annotations = self.findChild(QCheckBox, 'checkBox')
        self.checkBox_2 = self.findChild(QCheckBox,'checkBox_2')

        self.checkbox_annotations.setChecked(self.annotations)
        self.checkBox_2.setChecked(self.purge_widget)
        self.checkbox_annotations.stateChanged.connect(self.update_parameters)
        self.checkBox_2.stateChanged.connect(self.update_parameters)
        # Data Management
        self.save_path = None
        self.data = None
        self.json = False

    def update_parameters(self):
        self.annotations = self.checkbox_annotations.isChecked()
        self.purge_widget=self.checkBox_2.isChecked()
        self.run()

    def save_file(self):
        self.error("")
        self.warning("")
        if os.path.isdir(self.save_path):
            self.save_path = os.path.join(self.save_path, self.filename)

        import Orange.widgets.data.owsave as save_py
        saver = save_py.OWSave()
        saver.add_type_annotations = self.annotations
        filters = saver.valid_filters()
        extension = os.path.splitext(self.save_path)[1]
        selected_filter = ""
        for key in filters:
            if f"(*{extension})" in key:
                selected_filter = key
        if selected_filter == "":
            self.error(f"Invalid extension for savepath : {self.save_path}")
            return

        saver.data = self.data
        saver.filename = self.save_path
        saver.filter = selected_filter
        saver.do_save()
        self.Outputs.data.send(self.data)


    def save_json(self):
        if "content" not in self.data.domain:
            self.error("No answer column found.")
            return
        if "content" not in self.data.domain:
            self.error("No path column found.")
            return
        for i in range(len(self.data.get_column("path"))):
            text_response = self.data.get_column("content")[i]
            folder_path = self.data.get_column("path")[i]
            try:
                data_raw = json.loads(text_response)
            except json.JSONDecodeError as e:
                print("JSON mal formé :", e)
                try:
                    data_raw = ast.literal_eval(text_response)
                except Exception as e2:
                    print("Invalid JSON :", e2)
                    self.error("Invalid JSON :", e2)
                    return

            with open(folder_path, "w", encoding="utf-8") as f:
                json.dump(data_raw, f, ensure_ascii=False, indent=4)
        self.information("JSON saved successfully")
        self.Outputs.data.send(self.data)

    def run(self):
        self.error("")
        self.information("")
        """Save data to a file."""
        if self.data is None:
            self.error("need data")
            return
        if self.save_path is None:
            self.error("need path")
            return
        if self.json:
            self.save_json()
        else:
            self.save_file()

        if self.purge_widget:
            self.save_path = None
        to_send=copy.deepcopy(self.data)
        if self.purge_widget:
            self.data = None

        self.Outputs.data.send(to_send)




if __name__ == "__main__": 
    WidgetPreview(OWSaveFilepathEntry).run()



# fix error explained in https://github.com/pytorch/pytorch/issues/131662
import os
import sys


def fix_error_torch():
    """
    deprecated
    """
    return
    # if ptyhon!=3.11 ignore
    if sys.version_info[0]!=3:
        return
    if sys.version_info[1] != 11:
        return
    if sys.platform!="win32":
        return
    dest=os.path.dirname(os.__file__).replace("\\","/")+"/site-packages/torch/lib/libomp140.x86_64.dll"
    if os.path.exists(dest):
        return
    source=os.path.dirname(__file__).replace("\\","/")+"/libomp140.x86_64.dll"
    with open(source, 'rb') as src, open(dest, 'wb') as dst:
        dst.write(src.read())
fix_error_torch()




import os
import re
from typing import Any, Iterable, List, Set
from pygments.formatters import HtmlFormatter
import markdown
import sys
# Dans la classe OWCorpusViewer, ajoutez les méthodes suivantes :
import numpy as np
from AnyQt.QtCore import (
    QAbstractListModel,
    QEvent,
    QItemSelection,
    QItemSelectionModel,
    QItemSelectionRange,
    QModelIndex,
    QSortFilterProxyModel,
    Qt,
    QUrl,
)
from AnyQt.QtWidgets import (
    QAbstractItemView,
    QApplication,
    QHeaderView,
    QListView,
    QSizePolicy,
    QSplitter,
    QTableView,
)
from Orange.data import Variable
from Orange.data.domain import Domain, filter_visible
from Orange.widgets import gui
from Orange.widgets.settings import ContextSetting, Setting, DomainContextHandler
from Orange.widgets.utils.annotated_data import create_annotated_table
from Orange.widgets.utils.concurrent import ConcurrentWidgetMixin, TaskState
from Orange.widgets.utils.itemmodels import DomainModel
from Orange.widgets.widget import Input, Msg, Output, OWWidget
from orangecanvas.gui.utils import disconnected
from orangewidget.utils.listview import ListViewSearch

from orangecontrib.text.corpus import Corpus

HTML = """
<!doctype html>
    <html>
    <head>
    <meta charset='utf-8'>
    <style>
    /* Votre CSS existant */
    table {{ border-collapse: collapse; }}
    mark {{ background: #FFCD28; }}

    tr > td {{
        padding-bottom: 3px;
        padding-top: 3px;
    }}

    body {{
        font-family: Helvetica;
        font-size: 10pt;
    }}

    .line {{ border-bottom: 1px solid #000; }}
    .separator {{ height: 5px; }}

    .variables {{
        vertical-align: top;
        padding-right: 10px;
    }}

    .content {{
        overflow-wrap: break-word;
        word-wrap: break-word;
        -ms-word-break: break-all;
        word-break: break-all;
        word-break: break-word;
        -ms-hyphens: auto;
        -moz-hyphens: auto;
        -webkit-hyphens: auto;
        hyphens: auto;
    }}

    .token {{
        padding: 3px;
        border: 1px #B0B0B0 solid;
        margin-right: 5px;
        margin-bottom: 5px;
        display: inline-block;
    }}

    img {{
        max-width: 100%;
    }}
    </style>
    {additional_css}
    </head>
    <body>
    {html_content}
    </body>
    </html>
    """

SEPARATOR = (
    '<tr class="line separator"><td/><td/></tr>'
    '<tr class="separator"><td/><td/></tr>'
)


def _count_matches(content: List[str], search_string: str, state: TaskState) -> int:
    """
    Count number of appears of any terms in search_string in content texts.

    Parameters
    ----------
    content
        List of texts where we count appearances
    search_string
        Strings that are searched in texts. This parameter has a format
        term1|term2|term3|...

    Returns
    -------
    Number of all matches of search_string in all texts in content list
    """
    matches = 0
    if search_string:
        regex = re.compile(search_string.strip("|"), re.IGNORECASE)
        for i, text in enumerate(content):
            matches += len(regex.findall(text))
            state.set_progress_value((i + 1) / len(content) * 100)
    return matches

def is_qt_offscreen():
    return os.environ.get('QT_QPA_PLATFORM') == 'offscreen'

class DocumentListModel(QAbstractListModel):
    """
    Custom model for listing documents. Using custom model since Onrage's
    pylistmodel is too slow for large number of documents
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__visible_data = []
        self.__filter_content = []

    def data(self, index: QModelIndex, role: int = Qt.DisplayRole) -> Any:
        if role == Qt.DisplayRole:
            return self.__visible_data[index.row()]
        elif role == Qt.UserRole:
            return self.__filter_content[index.row()]

    def rowCount(self, parent: QModelIndex = None, *args, **kwargs) -> int:
        return len(self.__visible_data)

    def setup_data(self, data: List[str], content: List[str]):
        self.beginResetModel()
        self.__visible_data = data
        self.__filter_content = content
        self.endResetModel()

    def update_filter_content(self, content: List[str]):
        assert len(content) == len(self.__visible_data)
        self.__filter_content = content

    def get_filter_content(self) -> List[str]:
        return self.__filter_content

    def clear(self):
        self.beginResetModel()
        self.__visible_data = []
        self.__filter_content = []
        self.endResetModel()


class DocumentsFilterProxyModel(QSortFilterProxyModel):
    """Filter model for documents list"""

    __regex = None

    def set_filter_string(self, filter_string: str):
        self.__regex = re.compile(filter_string.strip("|"), re.IGNORECASE)
        self.invalidateFilter()

    def filterAcceptsRow(self, source_row: int, source_parent: QModelIndex) -> bool:
        """Filter document that mathc the filter string"""
        if self.__regex is None:
            # filter is not defined yet - show all
            return True
        else:
            index = self.sourceModel().index(source_row, 0, source_parent)
            content = self.sourceModel().data(index, Qt.UserRole)
            res = self.__regex.search(content)
            return bool(res)


class DocumentTableView(QTableView):
    """TableView that disables unselecting all items"""

    def selectionCommand(
        self, index: QModelIndex, event: QEvent = None
    ) -> QItemSelectionModel.SelectionFlags:
        flags = super().selectionCommand(index, event)
        selmodel = self.selectionModel()
        if not index.isValid():  # Click on empty viewport; don't clear
            return QItemSelectionModel.NoUpdate
        if selmodel.isSelected(index):
            currsel = selmodel.selectedIndexes()
            if len(currsel) == 1 and index == currsel[0]:
                # Is the last selected index; do not deselect it
                return QItemSelectionModel.NoUpdate
        if (
            event is not None
            and event.type() == QEvent.MouseMove
            and flags & QItemSelectionModel.ToggleCurrent
        ):
            # Disable ctrl drag 'toggle'; can be made to deselect the last
            # index, would need to keep track of the current selection
            # (selectionModel does this but does not expose it)
            flags &= ~QItemSelectionModel.Toggle
            flags |= QItemSelectionModel.Select
        return flags


class VariableListViewSearch(ListViewSearch):
    """ListViewSearch that disables unselecting all items in the list"""

    def selectionCommand(
        self, index: QModelIndex, event: QEvent = None
    ) -> QItemSelectionModel.SelectionFlags:
        flags = super().selectionCommand(index, event)
        selmodel = self.selectionModel()
        if not index.isValid():  # Click on empty viewport; don't clear
            return QItemSelectionModel.NoUpdate
        if selmodel.isSelected(index):
            currsel = selmodel.selectedIndexes()
            if len(currsel) == 1 and index == currsel[0]:
                # Is the last selected index; do not deselect it
                return QItemSelectionModel.NoUpdate
        if (
            event is not None
            and event.type() == QEvent.MouseMove
            and flags & QItemSelectionModel.ToggleCurrent
        ):
            # Disable ctrl drag 'toggle'; can be made to deselect the last
            # index, would need to keep track of the current selection
            # (selectionModel does this but does not expose it)
            flags &= ~QItemSelectionModel.Toggle
            flags |= QItemSelectionModel.Select
        return flags

    def set_selection(self, items: Iterable[Variable]):
        """Set selected items in the list view"""
        model = self.model()
        values = self.model()[:]
        items = [it for it in items if it in values]
        selection = QItemSelection()
        if items:
            for val in items:
                index = values.index(val)
                selection.merge(
                    QItemSelection(model.index(index, 0), model.index(index, 0)),
                    QItemSelectionModel.Select,
                )
        self.selectionModel().select(selection, QItemSelectionModel.ClearAndSelect)


class VisibleDomainModel(DomainModel):
    """Domain model that filter only visible features"""

    def set_domain(self, domain):
        if domain is not None:
            domain = Domain(
                filter_visible(domain.attributes),
                class_vars=filter_visible(domain.class_vars),
                metas=filter_visible(domain.metas),
            )
        super().set_domain(domain)


class OWCorpusViewer(OWWidget, ConcurrentWidgetMixin):
    name = "MarkDown Viewer"
    description = "Display corpus contents with Markdown."
    icon = "icons/MDViewer.png"
    category = "AAIT - TOOLBOX"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/MDViewer.png"
    priority = 1002
    keywords = "Markdown viewer, display"

    class Inputs:
        corpus = Input("Corpus", Corpus, replaces=["Data"])

    class Outputs:
        matching_docs = Output("Matching Docs", Corpus, default=True)
        other_docs = Output("Other Docs", Corpus)
        corpus = Output("Corpus", Corpus)

    settingsHandler = DomainContextHandler()

    settings_version = 2
    search_features: List[Variable] = ContextSetting([])
    display_features: List[Variable] = ContextSetting([])
    selected_documents: Set[int] = Setting({0}, schema_only=True)
    regexp_filter = ContextSetting("")
    show_tokens = Setting(False)
    autocommit = Setting(True)

    class Warning(OWWidget.Warning):
        no_feats_search = Msg("No features included in search.")
        no_feats_display = Msg("No features selected for display.")

    def __init__(self):
        super().__init__()
        ConcurrentWidgetMixin.__init__(self)

        self.corpus = None  # Corpus
        self.__pending_selected_documents = self.selected_documents

        # Info attributes
        self.update_info()
        info_box = gui.widgetBox(self.controlArea, "Info")
        gui.label(info_box, self, "Tokens: %(n_tokens)s")
        gui.label(info_box, self, "Types: %(n_types)s")
        gui.label(info_box, self, "Matching documents: %(n_matching)s")
        gui.label(info_box, self, "Matches: %(n_matches)s")

        # Search features
        ex_sel = QListView.ExtendedSelection
        search_box = gui.widgetBox(self.controlArea, "Search features")
        self.search_listbox = sl = VariableListViewSearch(selectionMode=ex_sel)
        search_box.layout().addWidget(sl)
        sl.setModel(VisibleDomainModel(separators=False))
        sl.selectionModel().selectionChanged.connect(self.search_features_changed)

        # Display features
        display_box = gui.widgetBox(self.controlArea, "Display features")
        self.display_listbox = dl = VariableListViewSearch(selectionMode=ex_sel)
        display_box.layout().addWidget(dl)
        dl.setModel(VisibleDomainModel(separators=False))
        dl.selectionModel().selectionChanged.connect(self.display_features_changed)

        self.show_tokens_checkbox = gui.checkBox(
            display_box,
            self,
            "show_tokens",
            "Show Tokens && Tags",
            callback=self.show_docs,
        )

        # Auto-commit box
        gui.auto_commit(
            self.controlArea, self, "autocommit", "Send data", "Auto send is on"
        )

        # Search
        self.filter_input = gui.lineEdit(
            self.mainArea,
            self,
            "regexp_filter",
            orientation=Qt.Horizontal,
            sizePolicy=QSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Fixed),
            label="RegExp Filter:",
            callback=self.refresh_search,
        )

        # Main area
        self.splitter = QSplitter(orientation=Qt.Horizontal, childrenCollapsible=False)
        # Document list
        self.doc_list = DocumentTableView()
        self.doc_list.setSelectionBehavior(QTableView.SelectRows)
        self.doc_list.setSelectionMode(QTableView.ExtendedSelection)
        self.doc_list.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.doc_list.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)
        self.doc_list.horizontalHeader().setVisible(False)
        self.splitter.addWidget(self.doc_list)

        self.doc_list_model = DocumentListModel()
        proxy_model = DocumentsFilterProxyModel()
        proxy_model.setSourceModel(self.doc_list_model)
        self.doc_list.setModel(proxy_model)
        self.doc_list.selectionModel().selectionChanged.connect(self.selection_changed)
        # Document contents
        self.doc_webview = None
        if not is_qt_offscreen():
            self.doc_webview = gui.WebviewWidget(self.splitter, debug=False)
            self.mainArea.layout().addWidget(self.splitter)

    def copy_to_clipboard(self):
        if self.doc_webview is not None:
            text = self.doc_webview.selectedText()
            QApplication.clipboard().setText(text)
        else:
            QApplication.clipboard().setText("")

    @Inputs.corpus
    def set_data(self, corpus=None):
        self.closeContext()
        self.reset_widget()
        self.corpus = corpus
        if corpus is not None:
            self.setup_controls()
            self.openContext(self.corpus)
            self.doc_list.model().set_filter_string(self.regexp_filter)
            self.select_variables()
            self.list_docs()
            self.update_info()
            self.set_selection()
            self.show_docs()
        self.commit.now()

    def reset_widget(self):
        # Corpus
        self.corpus = None
        # Widgets
        self.search_listbox.model().set_domain(None)
        self.display_listbox.model().set_domain(None)
        self.filter_input.clear()
        self.update_info()
        # Models/vars
        self.doc_list_model.clear()
        # Warnings
        self.Warning.clear()
        # WebView
        if self.doc_webview is not None:
            self.doc_webview.setHtml("")

    def setup_controls(self):
        """Setup controls in control area"""
        domain = self.corpus.domain
        if not self.corpus.has_tokens():
            self.show_tokens_checkbox.setCheckState(Qt.Unchecked)
        self.show_tokens_checkbox.setEnabled(self.corpus.has_tokens())
        self.search_listbox.model().set_domain(domain)
        self.display_listbox.model().set_domain(domain)
        self.search_features = self.search_listbox.model()[:]
        self.display_features = self.display_listbox.model()[:]

    def select_variables(self):
        """Set selection to display and search features view boxes"""
        smodel = self.search_listbox.model()
        dmodel = self.display_listbox.model()
        # it can happen that domain handler will set some features that are
        # not part of domain - remove them
        self.search_features = [f for f in self.search_features if f in smodel]
        self.display_features = [f for f in self.display_features if f in dmodel]
        # if no features after removing non-existent, select all - default
        if not self.search_features:
            self.search_features = smodel[:]
        if not self.display_features:
            self.display_features = dmodel[:]
        with disconnected(
            self.search_listbox.selectionModel().selectionChanged,
            self.search_features_changed,
        ):
            self.search_listbox.set_selection(self.search_features)
        with disconnected(
            self.display_listbox.selectionModel().selectionChanged,
            self.display_features_changed,
        ):
            self.display_listbox.set_selection(self.display_features)

    def list_docs(self):
        """List documents into the left scrolling area"""
        docs = self.regenerate_docs()
        self.doc_list_model.setup_data(self.corpus.titles.tolist(), docs)

    def get_selected_indexes(self) -> Set[int]:
        m = self.doc_list.model().mapToSource
        return {m(i).row() for i in self.doc_list.selectionModel().selectedRows()}

    def set_selection(self) -> None:
        """
        Select documents in selected_documents attribute in the view
        """
        self.selected_documents = self.__pending_selected_documents
        self.__pending_selected_documents = {0}
        view = self.doc_list
        model = view.model()
        source_model = model.sourceModel()

        selection = QItemSelection()
        self.selected_documents = {
            r for r in self.selected_documents if r < len(self.corpus)
        }
        for row in self.selected_documents:
            index = model.mapFromSource(source_model.index(row, 0))
            selection.append(QItemSelectionRange(index, index))
        # don't emit selection change to avoid double call of commit function
        # it is already called from set_data
        with disconnected(
            self.doc_list.selectionModel().selectionChanged, self.selection_changed
        ):
            view.selectionModel().select(selection, QItemSelectionModel.ClearAndSelect)

    def selection_changed(self) -> None:
        """Function is called every time the selection changes"""
        self.selected_documents = self.get_selected_indexes()
        self.show_docs()
        self.commit.deferred()


    def __mark_text(self, text):
        search_keyword = self.regexp_filter.strip("|")
        if not search_keyword:
            return text

        try:
            reg = re.compile(search_keyword, re.IGNORECASE | re.MULTILINE)
        except re.error:
            return text

        matches = list(reg.finditer(text))
        if not matches:
            return text

        text = list(text)
        for m in matches[::-1]:
            text[m.start() : m.end()] = list(
                f'<mark data-markjs="true">{"".join(text[m.start(): m.end()])}</mark>'
            )
        return "".join(text)

    @staticmethod
    def __get_selected_rows(view: QListView) -> List[Variable]:
        rows = view.selectionModel().selectedRows()
        values = view.model()[:]
        return [values[row.row()] for row in sorted(rows, key=lambda r: r.row())]

    def search_features_changed(self):
        self.search_features = self.__get_selected_rows(self.search_listbox)
        if self.corpus:
            self.doc_list_model.update_filter_content(self.regenerate_docs())
        self.doc_list.model().invalidateFilter()
        self.refresh_search()

    def display_features_changed(self):
        self.display_features = self.__get_selected_rows(self.display_listbox)
        self.show_docs()

    def regenerate_docs(self) -> List[str]:
        self.Warning.no_feats_search.clear()
        if len(self.search_features) == 0:
            self.Warning.no_feats_search()
        return self.corpus.documents_from_features(self.search_features)

    def refresh_search(self):
        if self.corpus is not None:
            self.doc_list.model().set_filter_string(self.regexp_filter)
            if not self.selected_documents:
                # when currently selected items are filtered selection is empty
                # select first element in the view in that case
                self.doc_list.setCurrentIndex(self.doc_list.model().index(0, 0))
            self.update_info()
            self.start(
                _count_matches,
                self.doc_list_model.get_filter_content(),
                self.regexp_filter,
            )
            self.show_docs()
            self.commit.deferred()

    def on_done(self, res: int):
        """When matches count is done show the result in the label"""
        self.n_matches = res if res is not None else "n/a"

    def on_exception(self, ex):
        raise ex

    def update_info(self):
        if self.corpus is not None:
            has_tokens = self.corpus.has_tokens()
            self.n_matching = f"{self.doc_list.model().rowCount()}/{len(self.corpus)}"
            self.n_tokens = self.corpus.count_tokens() if has_tokens else "n/a"
            self.n_types = self.corpus.count_unique_tokens() if has_tokens else "n/a"
        else:
            self.n_matching = "n/a"
            self.n_matches = "n/a"
            self.n_tokens = "n/a"
            self.n_types = "n/a"

    @gui.deferred
    def commit(self):
        matched = unmatched = annotated_corpus = None
        if self.corpus is not None:
            selected_docs = sorted(self.get_selected_indexes())
            matched = self.corpus[selected_docs] if selected_docs else None
            mask = np.ones(len(self.corpus), bool)
            mask[selected_docs] = 0
            unmatched = self.corpus[mask] if mask.any() else None
            annotated_corpus = create_annotated_table(self.corpus, selected_docs)
        self.Outputs.matching_docs.send(matched)
        self.Outputs.other_docs.send(unmatched)
        self.Outputs.corpus.send(annotated_corpus)

    def send_report(self):
        self.report_items(
            (
                ("Query", self.regexp_filter),
                ("Matching documents", self.n_matching),
                ("Matches", self.n_matches),
            )
        )

    def showEvent(self, event):
        super().showEvent(event)
        self.update_splitter()

    def update_splitter(self):
        """
        Update splitter that document list on the left never take more
        than 1/3 of the space. It is only set on showEvent. If user
        later changes sizes it stays as it is.
        """
        w1, w2 = self.splitter.sizes()
        ws = w1 + w2
        if w2 < 2 / 3 * ws:
            self.splitter.setSizes([int(ws * 1 / 3), int(ws * 2 / 3)])

    @classmethod
    def migrate_context(cls, context, version):
        if version < 2:
            f_order = context.values.pop("display_features", None)
            display_idx = context.values.pop("display_indices", [])
            search_ids = context.values.pop("search_indices", [])
            if f_order is not None:
                f_order = f_order[0]
                display_features = [f_order[i] for i in display_idx if i < len(f_order)]
                search_features = [f_order[i] for i in search_ids if i < len(f_order)]
                context.values["display_features"] = (display_features, -3)
                context.values["search_features"] = (search_features, -3)

            # old widget used PerfectDomainContextHandler with MATCH_VALUES_ALL
            # now it uses DomainContextHandler. The difference are:
            # - perfect handler stores values in tuple while domain in dicts
            # - domain context handler store class_vars together with attributes
            #   while perfect handler store them separately
            # - since MATCH_VALUES_ALL was used discrete var values were stored
            #   with var name (replacing them with id for discrete var - 1)
            if hasattr(context, "class_vars"):
                context.attributes = {
                    attr: 1 if isinstance(v, list) else v
                    for attr, v in context.attributes + context.class_vars
                }
                context.metas = dict(context.metas)
                delattr(context, "class_vars")

    def convert_markdown_with_syntax_highlighting(self, markdown_text, base_path=""):
        """
        Convertit le texte Markdown avec syntaxe de code colorée en HTML.
        """

        # --- Patch : normaliser les listes pour que les puces après un "1." restent des sous-listes ---
        markdown_text = self._normalize_lists_for_markdown(markdown_text)
        # ---------------------------------------------------------------------

        # Convertir le Markdown en HTML avec extensions pour le code
        html_content = markdown.markdown(
            markdown_text,
            extensions=['fenced_code', 'codehilite', 'tables', 'toc', 'md_in_html'],
            extension_configs={
                'codehilite': {
                    'guess_lang': False,
                    'pygments_style': 'default',
                    'noclasses': False  # Utiliser les classes CSS
                }
            }
        )

        # Ajouter le CSS pour Pygments
        css = HtmlFormatter().get_style_defs('.codehilite')
        html_content = f"<style>{css}</style>{html_content}"

        # Corriger les chemins d'images pour les fichiers locaux
        html_content = self.adjust_image_paths(html_content, base_path)
        return html_content

    def _normalize_lists_for_markdown(self, text: str) -> str:
        """
        Transforme les puces qui suivent immédiatement un item numéroté
        en sous-liste (en les indentant), pour éviter une renumérotation 1,2,3,4,5.

        Exemple :
            1. Titre
            2. Autre titre
            - sous-étape
            - sous-étape

        devient :
            1. Titre
            2. Autre titre
                - sous-étape
                - sous-étape
        """
        lines = text.splitlines()
        out = []
        in_ordered_list = False

        for line in lines:
            stripped = line.lstrip()

            # Ligne d'une liste numérotée : "1. xxx"
            if re.match(r'^\d+\.\s+', stripped):
                in_ordered_list = True
                # On normalise en supprimant l'indentation superflue éventuelle
                out.append(stripped)
                continue

            # Ligne d'une puce qui suit une liste numérotée
            if in_ordered_list and re.match(r'^[-*+]\s+', stripped):
                # Si elle n'est pas déjà indentée, on l'indente comme sous-liste
                if not line.startswith('    '):
                    out.append('    ' + stripped)
                else:
                    out.append(line)
                continue

            # Autre contenu : on sort du contexte "liste numérotée"
            in_ordered_list = False
            out.append(line)

        return "\n".join(out)

    def adjust_image_paths(self, html_content, base_path):
        """
        Corrige les chemins d'images pour qu'ils soient valides même pour des chemins relatifs.
        """

        def replace_path(match):
            image_path = match.group(1)

            # Si l'image est locale, corriger le chemin
            if not image_path.startswith(('http://', 'https://')):
                full_path = os.path.abspath(os.path.join(base_path, image_path))
                if os.path.exists(full_path):
                    print(f"[INFO] Image trouvée : {full_path}")
                    return f'src="file:///{full_path}"'
                else:
                    print(f"[ERREUR] Image introuvable : {full_path}")
                    return f'src="{image_path}"'  # Affiche toujours le chemin, même s'il est incorrect
            else:
                print(f"[INFO] Image distante : {image_path}")
                return f'src="{image_path}"'

        # Remplacer les chemins d'image
        adjusted_html = re.sub(r'src="([^"]+)"', replace_path, html_content)
        return adjusted_html

    def is_markdown(self, text):

        """Détecte si le texte est probablement du Markdown.

        Returns :
            bool : le text balisé
            """

        return bool(re.search(r'(```.+?```)|(`.+?`)', text, re.DOTALL))

    def get_pygments_css(self):
        """Génère le CSS pour Pygments.
                Returns :
                    str : le CSS de Pygments.
                    """
        formatter = HtmlFormatter(style='default')
        return f"<style>{formatter.get_style_defs()}</style>"
    def show_docs(self):
        """Affiche les documents sélectionnés dans la zone de droite.

        Affiche les documents stockés dans `selected_documents` dans la zone
        de droite. Pour cela, elle construit un tableau HTML avec les valeurs
        des différentes colonnes de la base de données.
        """
        if self.corpus is None:
            return

        # Afficher un message d'avertissement si on n'a pas de colonnes
        # sélectionnées pour l'affichage
        self.Warning.no_feats_display.clear()
        if len(self.display_features) == 0:
            self.Warning.no_feats_display()

        parts = []
        for doc_count, c_index in enumerate(sorted(self.selected_documents)):
            text = ""
            for feature in self.display_features:
                # Récupérer la valeur de la colonne
                value = str(self.corpus[c_index, feature.name])


                if feature.attributes.get("type", "") == "image" and value != "?":
                    value = os.path.join(feature.attributes.get("origin", ""), value)
                    value = f'<img src="{value}"></img>'

                else:
                    value = self.convert_markdown_with_syntax_highlighting(value, base_path="")


                # Ajouter la valeur formattée à la colonne
                text += (
                    f'<tr><td class="variables"><strong>{feature.name}:</strong></td>'
                    f'<td class="content">{value}</td></tr>'
                )

            # Ajouter la colonne formattée à la liste des colonnes
            parts.append(text)

        # Concatener les colonnes pour obtenir un tableau HTML
        joined = SEPARATOR.join(parts)

        # Générer le code HTML complet avec le style CSS et le tableau
        html_content = f"<table>{joined}</table>"
        css_styles = self.get_pygments_css()
        base = QUrl.fromLocalFile(__file__)
        if self.doc_webview is not None:
            self.doc_webview.setHtml(
                HTML.format(
                    additional_css=css_styles,
                    html_content=html_content
                ),
                base
            )

if __name__ == "__main__":
    app = QApplication(sys.argv)
    obj = OWCorpusViewer()
    obj.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()


# Copier-colle qui utilise la fonction de Windows (pour avoir la fenêtre classique qui s'ouvre)

import os
import threading
import time
if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\","/"):
    from Orange.widgets.orangecontrib.AAIT.utils import SimpleDialogQt
else:
    from orangecontrib.AAIT.utils import SimpleDialogQt
if os.name=='nt':
    import os.path

    from win32com.shell import shell, shellcon
    import ctypes
    import ctypes.wintypes
    from ctypes import windll, c_wchar_p
    import ctypes.wintypes as wt


    def win32_shellcopy(src, dest):
        """
        Copy files and directories using Windows shell.

        :param src: Path or a list of paths to copy. Filename portion of a path
                    (but not directory portion) can contain wildcards ``*`` and
                    ``?``.
        :param dst: destination directory.
        :returns: ``True`` if the operation completed successfully,
                  ``False`` if it was aborted by user (completed partially).
        :raises: ``WindowsError`` if anything went wrong. Typically, when source
                 file was not found.

        .. seealso:
            `SHFileperation on MSDN <http://msdn.microsoft.com/en-us/library/windows/desktop/bb762164(v=vs.85).aspx>`
        """
        if not os.path.exists(src):
            SimpleDialogQt.BoxError("Error file not exist : "+src)
            return
        if isinstance(src, str):
            src = os.path.abspath(src)
        else:  # iterable
            src = '\0'.join(os.path.abspath(path) for path in src)

        result, aborted = shell.SHFileOperation((
            0,
            shellcon.FO_COPY,
            src,
            os.path.abspath(dest),
            shellcon.FOF_NOCONFIRMMKDIR,  # flags
            None,
            None))

        if not aborted and result != 0:
            # Note: raising a WindowsError with correct error code is quite
            # difficult due to SHFileOperation historical idiosyncrasies.
            # Therefore we simply pass a message.
            raise WindowsError('SHFileOperation failed: 0x%08x' % result)

        return not aborted


    def select_folder_ctypes(title="Select a directory"):
        """
        Ouvre une boîte de sélection de dossier Windows toujours au premier plan
        (devant toutes les fenêtres du système), sans threads ni repop.
        """

        BIF_RETURNONLYFSDIRS = 0x0001
        BIF_NEWDIALOGSTYLE = 0x0040
        MAX_PATH = 32767

        BFFM_INITIALIZED = 1
        SWP_NOMOVE = 0x0002
        SWP_NOSIZE = 0x0001
        SWP_SHOWWINDOW = 0x0040
        HWND_TOPMOST = -1

        SW_SHOW = 5

        user32 = ctypes.windll.user32
        shell32 = ctypes.windll.shell32
        ole32 = ctypes.windll.ole32

        class BROWSEINFO(ctypes.Structure):
            _fields_ = [
                ("hwndOwner", wt.HWND),
                ("pidlRoot", ctypes.c_void_p),
                ("pszDisplayName", ctypes.c_wchar_p),
                ("lpszTitle", ctypes.c_wchar_p),
                ("ulFlags", wt.UINT),
                ("lpfn", ctypes.c_void_p),
                ("lParam", ctypes.c_void_p),
                ("iImage", wt.INT),
            ]

        BFFCALLBACK = ctypes.WINFUNCTYPE(wt.INT, wt.HWND, wt.UINT, wt.LPARAM, wt.LPARAM)

        SHBrowseForFolderW = shell32.SHBrowseForFolderW
        SHGetPathFromIDListW = shell32.SHGetPathFromIDListW
        CoTaskMemFree = ole32.CoTaskMemFree
        CoInitializeEx = ole32.CoInitializeEx
        CoUninitialize = ole32.CoUninitialize

        SHBrowseForFolderW.argtypes = [ctypes.POINTER(BROWSEINFO)]
        SHBrowseForFolderW.restype = ctypes.c_void_p
        SHGetPathFromIDListW.argtypes = [ctypes.c_void_p, ctypes.c_wchar_p]
        SHGetPathFromIDListW.restype = wt.BOOL

        def _cb(hwnd, uMsg, lParam, lpData):
            if uMsg == BFFM_INITIALIZED:
                # Force la fenêtre au tout premier plan
                user32.ShowWindow(hwnd, SW_SHOW)
                user32.SetForegroundWindow(hwnd)
                user32.SetActiveWindow(hwnd)
                user32.BringWindowToTop(hwnd)
                user32.SetWindowPos(hwnd, HWND_TOPMOST, 0, 0, 0, 0,
                                    SWP_NOMOVE | SWP_NOSIZE | SWP_SHOWWINDOW)
            return 0

        cb = BFFCALLBACK(_cb)

        COINIT_APARTMENTTHREADED = 0x2
        CoInitializeEx(None, COINIT_APARTMENTTHREADED)

        display_name_buffer = ctypes.create_unicode_buffer(MAX_PATH)
        bi = BROWSEINFO()
        bi.hwndOwner = 0  # pas d’attachement à une autre fenêtre
        bi.pidlRoot = None
        bi.pszDisplayName = ctypes.cast(display_name_buffer, ctypes.c_wchar_p)
        bi.lpszTitle = title
        bi.ulFlags = BIF_RETURNONLYFSDIRS | BIF_NEWDIALOGSTYLE
        bi.lpfn = ctypes.cast(cb, ctypes.c_void_p)
        bi.lParam = None
        bi.iImage = 0

        pidl = SHBrowseForFolderW(ctypes.byref(bi))

        folder_path = ""
        if pidl:
            out_buf = ctypes.create_unicode_buffer(MAX_PATH)
            if SHGetPathFromIDListW(pidl, out_buf):
                folder_path = out_buf.value
            CoTaskMemFree(ctypes.c_void_p(pidl))

        CoUninitialize()
        return folder_path.replace("\\", "/") if folder_path else ""


    # Sélectionner un fichier image unique
    # path = select_file_ctypes(multi_select=False, file_filter="Images (*.jpg;*.png;*.jpeg)")
    #
    # # Sélectionner plusieurs fichiers Word
    # paths = select_file_ctypes(multi_select=True, file_filter="Word Documents (*.docx;*.doc)")
    import ctypes
    import ctypes.wintypes


    def select_file_ctypes(multi_select=False, file_filter="All Files (*.*)", dialog_title="Select File(s)"):
        """
        Opens a native Windows file selection dialog with enforced modal-like behavior.

        Args:
            multi_select (bool): Allow selection of multiple files if True.
            file_filter (str): File type filters, e.g., "Images (*.jpg;*.png;*.jpeg)" or "Documents (*.docx;*.doc)".

        Returns:
            str: Selected file path(s) separated by ';', or an empty string if no selection was made.
        """
        OFN_ALLOWMULTISELECT = 0x00000200
        OFN_EXPLORER = 0x00080000
        OFN_FILEMUSTEXIST = 0x00001000
        MAX_PATH = 65536  # Large enough for multiple files

        class OPENFILENAME(ctypes.Structure):
            _fields_ = [
                ("lStructSize", ctypes.wintypes.DWORD),
                ("hwndOwner", ctypes.wintypes.HWND),
                ("hInstance", ctypes.wintypes.HINSTANCE),
                ("lpstrFilter", ctypes.c_wchar_p),
                ("lpstrCustomFilter", ctypes.c_wchar_p),
                ("nMaxCustFilter", ctypes.wintypes.DWORD),
                ("nFilterIndex", ctypes.wintypes.DWORD),
                ("lpstrFile", ctypes.c_wchar_p),
                ("nMaxFile", ctypes.wintypes.DWORD),
                ("lpstrFileTitle", ctypes.c_wchar_p),
                ("nMaxFileTitle", ctypes.wintypes.DWORD),
                ("lpstrInitialDir", ctypes.c_wchar_p),
                ("lpstrTitle", ctypes.c_wchar_p),
                ("Flags", ctypes.wintypes.DWORD),
                ("nFileOffset", ctypes.wintypes.WORD),
                ("nFileExtension", ctypes.wintypes.WORD),
                ("lpstrDefExt", ctypes.c_wchar_p),
                ("lCustData", ctypes.wintypes.LPARAM),
                ("lpfnHook", ctypes.c_void_p),
                ("lpTemplateName", ctypes.c_wchar_p),
                ("pvReserved", ctypes.c_void_p),
                ("dwReserved", ctypes.wintypes.DWORD),
                ("FlagsEx", ctypes.wintypes.DWORD),
            ]

        user32 = ctypes.windll.user32
        comdlg32 = ctypes.windll.comdlg32

        # Create filter string (must be null-separated and end with double null)
        if "(" in file_filter and ")" in file_filter:
            description = file_filter.split("(")[0].strip()
            pattern = file_filter[file_filter.find("(") + 1:file_filter.find(")")]
        else:
            description = "All Files"
            pattern = "*.*"
        filter_combined = f"{description}\0{pattern}\0All Files\0*.*\0\0"

        # Prepare buffer
        file_buffer = ctypes.create_unicode_buffer(MAX_PATH)

        ofn = OPENFILENAME()
        ofn.lStructSize = ctypes.sizeof(OPENFILENAME)
        ofn.hwndOwner = 0
        ofn.lpstrFilter = filter_combined
        ofn.lpstrFile = ctypes.cast(file_buffer, ctypes.c_wchar_p)
        ofn.nMaxFile = MAX_PATH
        ofn.lpstrTitle = dialog_title
        ofn.Flags = OFN_EXPLORER | OFN_FILEMUSTEXIST
        if multi_select:
            ofn.Flags |= OFN_ALLOWMULTISELECT

        # Avant ouvrir, mémoriser fenêtres existantes
        hwnd_list_before = []

        def enum_windows_proc(hwnd, lParam):
            hwnd_list_before.append(hwnd)
            return True

        EnumWindows = user32.EnumWindows
        EnumWindowsProc = ctypes.WINFUNCTYPE(ctypes.wintypes.BOOL, ctypes.wintypes.HWND, ctypes.wintypes.LPARAM)
        EnumWindows(EnumWindowsProc(enum_windows_proc), 0)

        # Ouvre la boîte de sélection
        if not comdlg32.GetOpenFileNameW(ctypes.byref(ofn)):
            return ""  # Cancelled

        time.sleep(0.1)

        # Après ouverture
        hwnd_list_after = []
        EnumWindows(EnumWindowsProc(lambda hwnd, lParam: hwnd_list_after.append(hwnd) or True), 0)

        new_hwnds = list(set(hwnd_list_after) - set(hwnd_list_before))

        selector_hwnd = None
        target_phrase = (dialog_title or "Select").lower()
        for hwnd in new_hwnds:
            length = user32.GetWindowTextLengthW(hwnd)
            if length > 0:
                window_title = ctypes.create_unicode_buffer(length + 1)
                user32.GetWindowTextW(hwnd, window_title, length + 1)
                if target_phrase in window_title.value.lower():
                    selector_hwnd = hwnd
                    break

        def enforce_modal(hwnd):
            if not hwnd:
                return
            while user32.IsWindow(hwnd):
                user32.SetForegroundWindow(hwnd)
                user32.SetWindowPos(hwnd, -1, 0, 0, 0, 0,
                                    0x0002 | 0x0001 | 0x0040)  # SWP_NOMOVE | SWP_NOSIZE | SWP_SHOWWINDOW
                time.sleep(0.2)

        if selector_hwnd:
            threading.Thread(target=enforce_modal, args=(selector_hwnd,), daemon=True).start()

        # Lecture du buffer brut
        buffer_content = ctypes.wstring_at(ctypes.addressof(file_buffer), file_buffer._length_)

        # Split sur \0
        parts = buffer_content.split('\0')
        parts = [p.replace("\\", "/") for p in parts if p]
        if multi_select:
            if len(parts) >= 2:
                folder = parts[0]
                files = parts[1:]
                full_paths = [f"{folder}/{file}" for file in files]
                return '//////'.join(full_paths)
            elif len(parts) == 1:
                return parts[0]
            else:
                return ""
        else:
            return parts[0] if parts else ""




    def select_new_file_ctypes(file_filter="All Files (*.*)", dialog_title="Create New File"):
        """
        Opens a native Windows file dialog for creating a new file.

        Args:
            file_filter (str): File type filters, e.g., "Text Files (*.txt)" or "All Files (*.*)".

        Returns:
            str: Selected file path as a string, or an empty string if cancelled.
        """
        OFN_EXPLORER = 0x00080000
        MAX_PATH = 65536  # Large buffer for filename

        class OPENFILENAME(ctypes.Structure):
            _fields_ = [
                ("lStructSize", ctypes.wintypes.DWORD),
                ("hwndOwner", ctypes.wintypes.HWND),
                ("hInstance", ctypes.wintypes.HINSTANCE),
                ("lpstrFilter", ctypes.c_wchar_p),
                ("lpstrCustomFilter", ctypes.c_wchar_p),
                ("nMaxCustFilter", ctypes.wintypes.DWORD),
                ("nFilterIndex", ctypes.wintypes.DWORD),
                ("lpstrFile", ctypes.c_wchar_p),
                ("nMaxFile", ctypes.wintypes.DWORD),
                ("lpstrFileTitle", ctypes.c_wchar_p),
                ("nMaxFileTitle", ctypes.wintypes.DWORD),
                ("lpstrInitialDir", ctypes.c_wchar_p),
                ("lpstrTitle", ctypes.c_wchar_p),
                ("Flags", ctypes.wintypes.DWORD),
                ("nFileOffset", ctypes.wintypes.WORD),
                ("nFileExtension", ctypes.wintypes.WORD),
                ("lpstrDefExt", ctypes.c_wchar_p),
                ("lCustData", ctypes.wintypes.LPARAM),
                ("lpfnHook", ctypes.c_void_p),
                ("lpTemplateName", ctypes.c_wchar_p),
                ("pvReserved", ctypes.c_void_p),
                ("dwReserved", ctypes.wintypes.DWORD),
                ("FlagsEx", ctypes.wintypes.DWORD),
            ]

        user32 = ctypes.windll.user32
        comdlg32 = ctypes.windll.comdlg32

        # Convert file_filter into Windows format: "Description\0Pattern\0..."
        if "(" in file_filter and ")" in file_filter:
            description = file_filter.split("(")[0].strip()
            pattern = file_filter[file_filter.find("(")+1:file_filter.find(")")]
        else:
            description = "All Files"
            pattern = "*.*"

        # Extract first extension for default
        first_ext = pattern.split(";")[0].strip().lstrip("*.")
        if not first_ext:
            first_ext = "txt"

        filter_combined = f"{description}\0{pattern}\0All Files\0*.*\0\0"

        file_buffer = ctypes.create_unicode_buffer(MAX_PATH)

        ofn = OPENFILENAME()
        ofn.lStructSize = ctypes.sizeof(OPENFILENAME)
        ofn.hwndOwner = 0
        ofn.lpstrFilter = filter_combined
        ofn.lpstrFile = ctypes.cast(file_buffer, ctypes.c_wchar_p)
        ofn.nMaxFile = MAX_PATH
        ofn.lpstrTitle = dialog_title
        ofn.Flags = OFN_EXPLORER
        ofn.lpstrDefExt = first_ext  # <- extension ajoutée automatiquement si absente

        # Mémoriser les fenêtres existantes
        hwnd_list_before = []
        def enum_windows_proc(hwnd, lParam):
            hwnd_list_before.append(hwnd)
            return True

        EnumWindows = user32.EnumWindows
        EnumWindowsProc = ctypes.WINFUNCTYPE(ctypes.wintypes.BOOL, ctypes.wintypes.HWND, ctypes.wintypes.LPARAM)
        EnumWindows(EnumWindowsProc(enum_windows_proc), 0)

        # Ouvre la boîte de dialogue
        if not comdlg32.GetSaveFileNameW(ctypes.byref(ofn)):
            return ""

        time.sleep(0.1)

        # Fenêtres après
        hwnd_list_after = []
        EnumWindows(EnumWindowsProc(lambda hwnd, lParam: hwnd_list_after.append(hwnd) or True), 0)
        new_hwnds = list(set(hwnd_list_after) - set(hwnd_list_before))

        selector_hwnd = None
        target_phrase = (dialog_title or "Save").lower()
        for hwnd in new_hwnds:
            length = user32.GetWindowTextLengthW(hwnd)
            if length > 0:
                window_title = ctypes.create_unicode_buffer(length + 1)
                user32.GetWindowTextW(hwnd, window_title, length + 1)
                if target_phrase in window_title.value.lower():
                    selector_hwnd = hwnd
                    break

        def enforce_modal(hwnd):
            if not hwnd:
                return
            while user32.IsWindow(hwnd):
                user32.SetForegroundWindow(hwnd)
                user32.SetWindowPos(hwnd, -1, 0, 0, 0, 0,
                                    0x0002 | 0x0001 | 0x0040)  # SWP_NOMOVE | SWP_NOSIZE | SWP_SHOWWINDOW
                time.sleep(0.2)

        if selector_hwnd:
            threading.Thread(target=enforce_modal, args=(selector_hwnd,), daemon=True).start()

        result = ctypes.wstring_at(ctypes.addressof(file_buffer), file_buffer._length_)
        return result.split('\0', 1)[0].replace("\\", "/")


    def BoxMessage(text: str, mode: int = 0, title: str = None) -> int:
        """
        Affiche une boîte native Windows (via ctypes).
        mode = 0 -> Error
        mode = 1 -> Warning
        mode = 2 -> Info
        Retourne l'ID du bouton cliqué (IDOK = 1).
        usage     BoxMessage("Ceci est une erreur", 0)
        BoxMessage("Ceci est un avertissement", 1)
        BoxMessage("Ceci est une information", 2)
        BoxMessage("Ceci est une erreur", 0,"titr1")
        BoxMessage("Ceci est un avertissement", 1,"titr2")
        BoxMessage("Ceci est une information", 2,"titr3")


        """


        # Flags boutons
        MB_OK = 0x00000000
        MB_SYSTEMMODAL   = 0x00001000
        MB_SETFOREGROUND = 0x00010000
        MB_TOPMOST       = 0x00040000

        # Icônes selon mode
        if mode == 0:  # Error
            icon_flag = 0x00000010  # MB_ICONERROR
            if title is None: title = "Error"
        elif mode == 1:  # Warning
            icon_flag = 0x00000030  # MB_ICONWARNING
            if title is None: title = "Warning"
        elif mode == 2:  # Info
            icon_flag = 0x00000040  # MB_ICONINFORMATION
            if title is None: title = "Information"
        else:
            raise ValueError("mode doit être 0=Error, 1=Warning, 2=Info")

        flags = MB_OK | icon_flag | MB_SYSTEMMODAL | MB_SETFOREGROUND | MB_TOPMOST

        return windll.user32.MessageBoxW(0, c_wchar_p(text), c_wchar_p(title), flags)




import ctypes  # to call id
import json
import os
import re
import shlex  # pour faire des split qui respectent les ""
import shutil
import subprocess
import sys
from urllib.parse import urlparse
import urllib.parse
from os.path import expanduser
from pathlib import Path
import time
import tempfile
import uuid
import base64
from Orange.data import Table, StringVariable, Domain
import platform



if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils import (SimpleDialogQt, aait_repo_file, shared_variables,
                     windows_utils,mac_utils)
else:
    from orangecontrib.AAIT.utils import (SimpleDialogQt, aait_repo_file, shared_variables,
                     windows_utils,mac_utils)


def force_save(self_instance):
    """
    This function forces the save of the currently opened Orange document.
    It utilizes the ctypes library to invoke the save_scheme function of the document.
    It checks if the save is successful and displays an error message if it fails.
    """

    # Initialize the canvas ID to a default value
    canvas_id = -666

    # Iterate through the shared variables to find the document's canvas ID
    for shared_variable in shared_variables.vect_doc_ptr:
        if self_instance.current_ows == shared_variable[0]:
            canvas_id = shared_variable[1]

    # Check if the canvas ID is still the default value, indicating failure to find the document
    if canvas_id == -666:
        print("Unable to save. Document not found.")
        return

    # Use ctypes to invoke the save_scheme function of the document
    save_result = ctypes.cast(canvas_id, ctypes.py_object).value.save_scheme()

    # Check the result of the save operation
    if save_result != 1:
        self_instance.error("Error: You need to save your document.")
        return

    # Clear any previous error messages
    self_instance.error("")


def force_save_as(self_instance, target_canvas_id=None):
    """
    This function forces the save of the currently opened Orange document with a new name. It loops until the save is successful.
    """
    save_result = 0

    # Loop until the save operation is successful
    while save_result != 1:
        self_instance.error("You need to save your document to use AAIT widgets")
        print("You need to save your document to use AAIT widgets")

        # Use ctypes to invoke the save_scheme function of the document
        save_result = ctypes.cast(target_canvas_id, ctypes.py_object).value.save_scheme()

    # Clear any previous error messages
    self_instance.error("")


def replaceAll(file_path, search_exp, replace_exp):
    """
    This function searches and replaces all occurrences of search_exp with replace_exp in a specified file.
    """
    try:
        modified_lines = []  # Create an empty list to store modified lines
        with open(file_path, "r") as file:
            lines = file.readlines()  # Read all lines from the file

        for line in lines:
            if search_exp in line:  # Check if the search expression is present in the line
                line = line.replace(search_exp,
                                    replace_exp)  # Replace the search expression with the replacement expression
            modified_lines.append(line)  # Append the modified line to the list

        print(modified_lines)  # Print the modified lines to the console

        with open(file_path, 'w') as file:
            file.writelines(modified_lines)  # Write the modified lines back to the file

    except Exception as e:
        print(e, file=sys.stderr)  # Print the exception details to standard error
        print("Error in modifying file:", file_path, file=sys.stderr)  # Print an error message to standard error
        return ""  # Return an empty string if there is an error


def current_met_directory(argself):
    """
    This function returns the current save directory for the current Orange document.
    It ensures that the document has been saved and that a save directory exists.
    """
    # Check if the current document is not saved or has an empty name
    if argself.current_ows == "toto" or argself.current_ows == "":
        print("Error: OWS file needs to be saved before using AAIT modules!", file=sys.stderr)

        # Force save the document with a new name
        force_save_as(argself)

        # Update the current document variable
        argself.current_ows = shared_variables.get_current_ows()

        # Recursive call to ensure a valid met directory is obtained
        return current_met_directory(argself)

    # Build the met directory path based on the current document path
    met_dir = argself.current_ows[:-3] + "metdir"

    # Check if the directory already exists
    if os.path.isdir(met_dir):
        return met_dir

    # Try creating the directory
    try:
        os.mkdir(met_dir)
    except Exception as e:
        print(e, file=sys.stderr)
        print("Error in creating directory:", met_dir, file=sys.stderr)
        return ""

    return met_dir


def write_met_file(self_arg, extra_arg_title=[], extra_arg_value=[]):
    """
    This function writes a file in "met" format (configuration) with specified titles and values.
    It checks for consistency between the lists of titles and values.
    """
    # Check for inconsistency in the length of title and value lists
    if len(extra_arg_title) != len(extra_arg_value):
        print("Error: Mismatch in length of arguments [title][value]", file=sys.stderr)
        return 1  # Return 1 to indicate an error

    # Get the current met directory
    met_dir = current_met_directory(self_arg)

    # Check if the met directory is not available
    if met_dir == "":
        # Show a save file window and call the function again
        print("Here, display a save file window and recall the function")
        return 1  # Return 1 to indicate an error

    # Construct the file name using the met directory and the caption title of self_arg
    file_name = met_dir + "/" + self_arg.captionTitle + ".met"

    try:
        with open(file_name, 'w') as file:
            # Write caption, name, and id to the file
            file.write("Caption ")
            file.write('"' + str(self_arg.captionTitle) + '"\n')
            file.write("Name ")
            file.write('"' + str(self_arg.name) + '"\n')
            file.write("Id ")
            file.write('"' + str(id(self_arg)) + '"')

            # Write additional titles and values to the file
            for i in range(len(extra_arg_title)):
                file.write("\n")
                file.write('"' + str(extra_arg_title[i]) + '" "')
                file.write(str(extra_arg_value[i]) + '"')

    except Exception as e:
        print(e, file=sys.stderr)
        print("Error in writing file:", file_name, file=sys.stderr)
        return 1  # Return 1 to indicate an error

    return 0  # Return 0 to indicate success


# 0 ok 1 erreur
def read_met_file_from_caption(self_instance, caption, out_title_list, out_value_list):
    """
    This function reads a "met" file based on a specified caption, extracts titles and values,
    and then stores them in the lists out_title_list and out_value_list.
    """
    # Get the current met directory
    met_directory = current_met_directory(self_instance)

    # Check if the met directory is not valid
    if met_directory == "":
        force_save_as(self_instance)
        # print("End of the attempt")

        # Update the current document variable
        self_instance.current_ows = shared_variables.get_current_ows()

        # Recursive call to ensure a valid met directory is obtained
        return read_met_file_from_caption(self_instance, caption, out_title_list, out_value_list)

    # Build the absolute path to the met file based on the caption
    absolute_path = met_directory + "/" + caption + ".met"

    # Call the function to read the met file from the absolute path
    return read_met_file_from_absolute_path(absolute_path, out_title_list, out_value_list)


def read_met_file_from_absolute_path(file_path, title_list, value_list):
    """
    This function reads a "met" file from its absolute path (file_path), extracts titles and values,
    and then stores them in the lists title_list and value_list.
    """
    lines_to_process = []
    del title_list[:]
    del value_list[:]

    try:
        with open(file_path, 'r') as file:
            lines_to_process = file.readlines()
    except Exception as e:
        print("Error in reading:", file_path, file=sys.stderr)
        print(e)
        return 1

    for i in range(0, len(lines_to_process)):
        cleaned_line = lines_to_process[i].strip()
        split_line = shlex.split(cleaned_line)

        if len(split_line) != 2:
            print("Line not expected:", lines_to_process[i], file=sys.stderr)
            return 1

        title_list.append(split_line[0])

        if len(split_line[1]) > 2:
            if (split_line[1][0] == '"' and split_line[1][-1] == '"'):
                split_line[1] = split_line[1][1:]
                split_line[1] = split_line[1][:-1]

        value_list.append(split_line[1])

    if len(title_list) != len(value_list):
        print("Error: Met file truncated size")
        return 1

    if len(title_list) < 3:
        print("Error: Met file truncated, less than 3 elements")
        return 1

    if title_list[0] != "Caption":
        print("Error: Caption not found in Met file at index 0")
        return 1

    if title_list[1] != "Name":
        print("Error: Name not found in Met file at index 1")
        return 1

    if title_list[2] != "Id":
        print("Error: Id not found in Met file at index 2")
        return 1

    return 0


def get_all_captions(self_instance, include_dict_variable=False):
    """
    This function returns the list of all caption names for "met" files in the current save directory.
    It allows including or excluding "dict_variable" files based on the include_dict_variable parameter.
    """
    # Get the current metm directory
    met_directory = current_met_directory(self_instance)

    # Check if the met directory is not valid
    if met_directory == "":
        print("Put a save file window here and call the function again")
        return 1

    # List to store file captions
    captions_list = []

    # Iterate through the directory
    for file_name in os.listdir(met_directory):
        # Check if the file is a "met" file
        if file_name.endswith('.met'):
            # Check if excluding "dict_variable" files and skip if necessary
            if not include_dict_variable and file_name[:-4] == "dict_variable":
                continue

            # Append the caption name to the list
            captions_list.append(file_name[:-4])

    return captions_list


def get_all_captions_with_specific_class(self_instance, class_name):
    """
    This function returns a list of caption names for "met" files that have a specified class (class_name).
    """
    # Get the list of all captions
    captions_to_study = get_all_captions(self_instance)

    # Check for an error in getting the list of captions
    if type(captions_to_study) is int:
        return []

    result_list = []

    # Check if the list of captions is empty
    if len(captions_to_study) == 0:
        return result_list

    # Iterate through the list of captions
    for caption_element in captions_to_study:
        title_list = []
        value_list = []

        # Read the "met" file and check for errors
        if 0 != read_met_file_from_caption(self_instance, caption_element, title_list, value_list):
            print("Error reading ", caption_element, file=sys.stderr)
            return []

        # Check if the class of the file matches the specified class_name
        if value_list[1] == class_name:
            result_list.append(caption_element)

    return result_list


def get_all_captions_from_specific_ows(ows_path, include_dict_variable=False):
    """
    This function returns the list of all caption names for "met" files in the save directory associated with a specified OWS document.
    """
    # Build the met directory path based on the OWS document path
    met_directory = ows_path[:-3] + "metdir"

    # List to store file captions
    captions_list = []

    # Iterate through the directory
    for file_name in os.listdir(met_directory):
        # Check if the file is a "met" file
        if file_name.endswith('.met'):
            # Check if excluding "dict_variable" files and skip if necessary
            if not include_dict_variable and file_name[:-4] == "dict_variable":
                continue

            # Append the caption name to the list
            captions_list.append(file_name[:-4])

    return captions_list


def get_all_captions_with_specific_class_from_specific_ows(class_name, ows_path):
    """
    This function returns a list of caption names for "met" files that have a specified class (class_name) in the save directory associated with a specified OWS document.
    """
    # Get the list of all captions for the specific OWS document
    captions_to_study = get_all_captions_from_specific_ows(ows_path)

    # Check for an error in getting the list of captions
    if type(captions_to_study) is int:
        return []

    result_list = []

    # Check if the list of captions is empty
    if len(captions_to_study) == 0:
        return result_list

    # Build the met directory path based on the OWS document path
    met_directory = ows_path[:-3] + "metdir"

    # Iterate through the list of captions
    for caption_element in captions_to_study:
        title_list = []
        value_list = []

        # Build the absolute path to the "met" file
        absolute_path = met_directory + "/" + caption_element + ".met"

        # Read the "met" file and check for errors
        if 0 != read_met_file_from_absolute_path(absolute_path, title_list, value_list):
            print("Error reading ", absolute_path, file=sys.stderr)
            return []

        # Check if the class of the file matches the specified class_name
        if value_list[1] == class_name:
            result_list.append(caption_element)

    return result_list


def is_caption_file_exist(self_arg, caption):
    """
    This function checks if a "met" file with a specific caption exists in the current save directory.
    """
    met_dir = current_met_directory(self_arg)

    # Check if the met_dir directory is not available
    if met_dir == "":
        # print("Here, display a save file window and recall the function")
        return 1  # Return 1 to indicate an error

    # Construct the absolute path using the met directory and the caption
    absolute_path = met_dir + "/" + caption + ".met"

    # Check if the file exists at the constructed absolute path
    return os.path.isfile(absolute_path)


def write_local_current_version():
    """"
    This function write local current version
    """
    version="0.0.0.0"# to be changed at each new version
    store_ia_path = expanduser("~")
    store_ia_path=store_ia_path.replace("\\","/")
    store_ia_path+="/aait_store/Parameters"
    # test for none standard path
    none_standard_path = sys.executable.replace("\\", "/")
    none_standard_path = os.path.dirname(none_standard_path)
    none_standard_path = os.path.dirname(none_standard_path)+"/aait_store/"
    if os.path.isfile(none_standard_path+"remote_ressources_path.txt"):
        store_ia_path=none_standard_path+"Parameters"


    version_file=store_ia_path+"/Store_IA.txt"
    try:
        os.makedirs(store_ia_path, exist_ok=True)
        with open(version_file, 'w') as f:
            f.write(version)
            pass
    except:
        SimpleDialogQt.BoxError("error impossible to write file :"+ version_file)
        raise Exception("error impossible to write file :"+ version_file)





def get_local_store_path():
    """
    This function return the IA Store path stocked locally on user computer
    create folder if not exist
    """
    store_ia_path = expanduser("~")
    store_ia_path = store_ia_path.replace("\\", "/")
    store_ia_path += "/aait_store/"

    # test for none standard path
    none_standard_path = sys.executable.replace("\\", "/")
    none_standard_path = os.path.dirname(none_standard_path)
    none_standard_path = os.path.dirname(none_standard_path)+"/aait_store/"




    if os.path.isfile(none_standard_path+"remote_ressources_path.txt"):
        store_ia_path=none_standard_path
    elif  os.path.isdir(none_standard_path):
        try:
            with open(none_standard_path + '/remote_ressources_path.txt', 'w'):
                pass
            store_ia_path=none_standard_path
        except:
            SimpleDialogQt.BoxError("error impossible to write file :" + none_standard_path)
            raise Exception("error impossible to write file :" + none_standard_path)

    if not os.path.exists(store_ia_path):
        # Create a new directory because it does not exist
        os.makedirs(store_ia_path,exist_ok=True)
        # create blanck file with future remote path
        try:
            with open(store_ia_path + '/remote_ressources_path.txt', 'w'):
                pass
        except:
            SimpleDialogQt.BoxError("error impossible to write file :"+ store_ia_path)
            raise Exception("error impossible to write file :"+ store_ia_path)
    return store_ia_path


def get_widget_extension_path():
    local_store_ia = get_local_store_path()
    result=""
    if os.path.exists(local_store_ia+"widget_extention_path.txt"):
        with open(local_store_ia+"widget_extention_path.txt", 'r', encoding='utf-8') as f:
            result=f.readline().strip()
            result.replace("\\", "/")
            if len(result)>1:
                if result[-1]!="/":
                    result=result+"/"
            return result
    if result=="":
        return local_store_ia+"AddOn/PythonExtension/"

def get_category_extension_to_load():
    the_widget_extention_path=get_widget_extension_path()
    if not os.path.isdir(the_widget_extention_path):
        return []
    widget_extention_to_check = [f for f in os.listdir(the_widget_extention_path) if os.path.isdir(os.path.join(the_widget_extention_path, f))]
    result=[]
    for element in widget_extention_to_check:
        path_to_check=the_widget_extention_path+element+"/__init__.py"
        if os.path.isfile(path_to_check):
            result.append(element)
    return result

def set_aait_store_remote_ressources_path(ressource_path):
    """
    Set up remote ressources path of store IA
    """
    if len(ressource_path) == 0:
        return
    ressource_path.replace("\\", "/")
    if is_url(ressource_path):
        path = urlparse(ressource_path).path
        if bool (os.path.splitext(path)[1])==False:
            if os.path.isfile(ressource_path)==False:
                if ressource_path[-1] != "/":
                    ressource_path += "/"

        else:
            if ressource_path.endswith('.aait'):
                temp_url = ressource_path.split("/")[:-1]
                ressource_path = "/".join(temp_url)+"/"

    else :
        if os.path.isfile(ressource_path) == False:
            if ressource_path[-1] != "/":
                ressource_path += "/"

    # check if proposed directory is valid
    try:
        version = get_aait_store_remote_version(ressource_path)
        print("current remote version", version)
    except:
        SimpleDialogQt.BoxError(ressource_path+ " is not a valid remote ressource path")
        print(ressource_path+ " is not a valid remote ressource path")
        return
    store_ia_path = get_local_store_path()
    # writinf local file with directory path
    try:
        with open(store_ia_path + '/remote_ressources_path.txt', 'w') as fp:
            fp.write(ressource_path)
            pass
    except:
        SimpleDialogQt.BoxError("error impossible to write file :"+ store_ia_path + '/remote_ressources_path.txt')
        raise Exception("error impossible to write file :"+ store_ia_path + '/remote_ressources_path.txt')


def get_aait_store_remote_version(ressource_path):
    """
    Return current remote version of store IA
    """
    if IsStoreCompressed(ressource_path):
        if get_size(ressource_path)==0:
            ressource_path=os.path.dirname(ressource_path)+"/"
        else:
            version=aait_repo_file.decode(ressource_path,"Parameters/Store_IA.txt")
            version=version.split(" ")[-1]
            return version
    if is_url(ressource_path):
        version_file = ressource_path + "/Parameters/Store_IA.txt"
    else:
        version_file = os.path.join(ressource_path, "Parameters/Store_IA.txt")

    version = ""
    try:
        if is_url(ressource_path):
            # Télécharger le contenu du fichier à distance
            curl_command = f"curl -s {version_file}"
            process = subprocess.run(curl_command, shell=True, capture_output=True, text=True)

            if process.returncode == 0:  # Vérifie si curl a réussi
                line = process.stdout.splitlines()[0]  # Lire la première ligne
                version = line.split(" ")[-1]
            else:
                raise Exception(f"Error: Unable to fetch file using curl: {version_file}")
        else:
            # Lire le fichier local
            with open(version_file, 'r') as f:
                line = f.readline()
                version = line.split(" ")[-1]
    except Exception as e:
        SimpleDialogQt.BoxError(f"Error: Impossible to read file: {version_file}\n{str(e)}")
        raise Exception(f"Error: Impossible to read file: {version_file}\n{str(e)}")

    print("Version:", version)
    return version


def get_aait_store_remote_ressources_path():
    """
    Get remote ressources path of store IA
    """
    try:
        local_aait_store_path = get_local_store_path()
    except Exception:
        SimpleDialogQt.BoxError("Impossible to open local ia store path")
        raise Exception("Impossible to open local ia store path")

    path_to_read = local_aait_store_path + "remote_ressources_path.txt"
    output_path = ""
    try:
        with open(path_to_read, 'r') as f:
            output_path = f.readline().strip()
            pass
    except:
        SimpleDialogQt.BoxError("error impossible to read file :"+ path_to_read)
        raise Exception("error impossible to read file :"+ path_to_read)
    return output_path



def evolved_json_load(json_data_str):
    """IMPORTANT: This feature is still in progress. If used, be aware that some bugs could happen."""
    print("Using evolved_json_load function. Feature still in progress")
    subcategories = {}
    resolved_data = []
    json_data = json.loads(json_data_str)

    print("json_data_type:", type(json_data))

    # Parse subcategories and create a mapping
    for entry in json_data:
        if "sous_category_name" in entry:
            subcategories[entry["sous_category_name"]] = {
                "workflows": entry.get("workflows", []),
                "extra_datas": entry.get("extra_datas", []),
            }

    # Resolve each entry that references subcategories
    for entry in json_data:
        if "use_sous_category" in entry:
            merged_workflows = entry.get("workflows", [])
            merged_extra_datas = entry.get("extra_datas", [])

            for subcat_name in entry["use_sous_category"]:
                if subcat_name in subcategories:
                    subcat = subcategories[subcat_name]
                    merged_workflows.extend(subcat["workflows"])
                    merged_extra_datas.extend(subcat["extra_datas"])

            # Remove duplicates
            entry["workflows"] = list(set(merged_workflows))
            entry["extra_datas"] = list(set(merged_extra_datas))

        # Append the resolved entry to the resolved_data list
        if "sous_category_name" not in entry:
            if isinstance(entry, str):
                print("WTF: ", entry)
            entry.pop("use_sous_category", None)
            resolved_data.append(entry)
    
    return resolved_data


def IsLocalStoreCompressed(aait_store_path):
    """
    Return the dictionnary of requierment
    """
    return not Path(aait_store_path).is_dir()

def is_url(path_str: str) -> bool:
    """Retourne True si la chaîne ressemble à une URL (http ou https)."""
    parsed = urllib.parse.urlparse(path_str)
    return parsed.scheme in ("http", "https")

def IsStoreCompressed(aait_store_path):
    """
    Determines if the given store path is considered "compressed."

    Logic:
      1) If it's a local path:
         - If it's a directory, return False (not compressed)
         - Else, return True (compressed)
      2) If it's a remote URL (HTTP/HTTPS):
         - HEAD request with curl. If Content-Type not HTML => probably a file => True
         - If HTML => we GET it to check for "Index of" or "Contenu de" => if found => False, else True
         - If we get 404, we do an extra GET to see if it might be a directory response or a custom 404
           - If we detect "Index of" or "Contenu de," we treat as directory => False
           - Otherwise, True
    """

    if not is_url(aait_store_path):
        return not Path(aait_store_path).is_dir()

    # ---------------------------------------------
    # CASE: The path is a remote HTTP/HTTPS URL
    # ---------------------------------------------
    process_head = subprocess.run(
        ["curl", "-s", "-L", "-I", aait_store_path],
        capture_output=True
    )
    if process_head.returncode != 0:
        print("La requête HEAD a échoué pour", aait_store_path)
        return True

    headers_str = process_head.stdout.decode("utf-8", errors="replace").lower()
    is_404 = "404 not found" in headers_str
    if not is_404:
        if "content-type: text/html" in headers_str:
            if has_listing_signature(aait_store_path):
                return False
            else:
                return True
        else:
            return True
    else:

        process_get = subprocess.run(
            ["curl", "-s", "-L", aait_store_path],
            capture_output=True
        )
        if process_get.returncode != 0:
            print("La requête GET a également échoué.")
            return True
        body = process_get.stdout.decode("utf-8", errors="replace").lower()

        if ("index of" in body) or ("contenu de" in body):
            #print("Listing de dossier détecté.")
            return False

        temp_url = aait_store_path.split("/")[:-1]
        url = "/".join(temp_url) + "/"
        process_get = subprocess.run(
            ["curl", "-s", "-L", url],
            capture_output=True
        )
        if process_get.returncode != 0:
            print("La requête GET a également échoué.")
            return True
        body = process_get.stdout.decode("utf-8", errors="replace").lower()
        lines = body.split("\n")
        for i in range(len(lines)):
            if ".aait" in lines[i]:
                match = re.search(r'(\d+)\s*octets', lines[i+2])
                if match:
                    if int(match.group(1)) < 10:
                        return False



        if ("index of" in body) or ("contenu de" in body):
            # print("Listing de dossier détecté.")
            return False
        return True


def has_listing_signature(url):
    """
    Does a quick GET on 'url' and checks if the body contains
    'Index of' or 'Contenu de'.
    Returns True if it likely is a directory listing,
    otherwise False.
    """
    process_get = subprocess.run(
        ["curl", "-s", "-L", url],
        capture_output=True
    )
    if process_get.returncode != 0:
        return False
    body = process_get.stdout.decode("utf-8", errors="replace").lower()
    return ("index of" in body) or ("contenu de" in body)

def get_aait_store_requirements_json(repo_path=None):
    """
    Return the dictionary of requirements
    :param repo_path: Optional repository path to use instead of the one from the file
    """
    try:
        if repo_path is None:
            aait_store_remote = get_aait_store_remote_ressources_path()
        else:
            aait_store_remote = repo_path
    except:
        SimpleDialogQt.BoxError("no access to ia store resource path")
        raise Exception("no access to ia store resource path")

    if not is_url(aait_store_remote):
        if IsLocalStoreCompressed(aait_store_remote):
            output = aait_repo_file.decode(aait_store_remote, "Parameters/requirements.json")
            if output == None:
                return
            data = evolved_json_load(output)
            return data
    # Cas distant (URL)
    json_path = aait_store_remote + "Parameters/requirements.json"
    if is_url(json_path):
        json_path = aait_store_remote + "/Parameters/requirements.json"
        # Utiliser curl pour récupérer le contenu du fichier distant
        curl_command = f"curl -s {json_path}"
        process = subprocess.run(curl_command, shell=True, capture_output=True, text=True)

        if process.returncode == 0:  # Vérifie si curl a réussi
            try:
                data = evolved_json_load(process.stdout)  # Utiliser stdout pour récupérer la sortie
                print("Données chargées depuis l'URL :", data)
                return data
            except Exception as e:
                SimpleDialogQt.BoxError(f"Error parsing JSON from URL: {json_path}")
                raise Exception(f"Error parsing JSON from URL: {json_path}") from e
        else:
            raise Exception(f"Error: Unable to fetch file using curl: {json_path}")
    else:
        if not Path(json_path).is_file():
            return {}
        try:
            with open(json_path, 'r', encoding='utf-8') as file:
                data = evolved_json_load(file.read())
                return data
        except:
            SimpleDialogQt.BoxError("unable to read"+ json_path)
            raise Exception("unable to read"+ json_path)


def download_aait_store_file_or_directory(server_path: str = "", target: str = "", repo_path: str = None) -> None:
    """
    copy a file/directory from remote to local
    :param server_path: Path to the server directory
    :param target: Target file or directory to download
    :param repo_path: Optional repository path to use instead of the one from the file
    """

    if server_path == "":
        return

    if server_path[-1] == "/":
        server_path = server_path[:-1]

    file_or_directory = server_path +"/"+ target


    try:
        if repo_path is None:
            aait_store_remote = get_aait_store_remote_ressources_path()
        else:
            aait_store_remote = repo_path
    except Exception as e:
        SimpleDialogQt.BoxError("no acces to ia store ressource path")
        print(e)
        raise Exception("no acces to ia store ressource path")

    compressed = IsStoreCompressed(aait_store_remote)

    # Helper function to check if a string is a URL
    def is_url(path_str):
        parsed = urllib.parse.urlparse(path_str)
        return parsed.scheme in ("http", "https")

    if not compressed:
        if not is_url(aait_store_remote):
            source =  file_or_directory
            output = get_local_store_path() + target

            if already_downloaded(source, output):
                return

            os.makedirs(os.path.dirname(output), exist_ok=True)

            if os.name == 'nt':
                # correction buf sinon on copie C:\toto\titi\ dans C:\tatat\titi\ on obtient C:\tatat\titi\titi ...
                if os.path.isdir(server_path):
                    output=str(Path(output).resolve().parent).replace('\\','/')+"/"
                windows_utils.win32_shellcopy(source, output)
            elif platform.system() == "Darwin":
                if os.path.isdir(server_path):
                    output=str(Path(output).resolve().parent).replace('\\','/')+"/"
                mac_utils.mac_copy_with_anyqt_progress(source, output)

            else:
                #try:
                #    windows_utils.mac_shellcopy(source, output)
                #except Exception as e:
                    #print(e)
                try:
                    dialog_progress = SimpleDialogQt.ProgressDialog(
                        title="Synchronsation please wait",
                        content=f"synchronization of the file {source}\nthis operation may take a few minutes"
                    )
                    dialog_progress.show()
                    if os.path.isfile(source):
                        shutil.copyfile(source, output)
                    else:
                        shutil.copytree(source, output,dirs_exist_ok=True)
                except Exception as e:
                    print(e)
                    raise Exception(f"erreur then copy{source}")

        else:
            # ----- REMOTE URL (non-compressed) -----


            try:
                aait_repo_file.download_from_folder_server(
                    base_url=aait_store_remote,
                    target_subfolder=target,
                    local_dir=get_local_store_path()
                )
            except Exception as e:
                print(e)
                SimpleDialogQt.BoxError(f"Error downloading file from URL: {aait_store_remote}")
                raise Exception(f"Error downloading file from URL: {aait_store_remote}")

        return

    # compressed case
    output = os.path.join(get_local_store_path(), target)
    if already_downloaded_compressed(aait_store_remote,file_or_directory,output):
        print(output, "already downloaded -> skip")
        return
    os.makedirs(os.path.dirname(output), exist_ok=True)

    #showing a freezed dialog box
    dialog_progress = SimpleDialogQt.ProgressDialog(title="Synchronsation please wait",content="synchronization of the file "+file_or_directory+"\nthis operation may take a few minutes")
    dialog_progress.show()
    aait_repo_file.decode_to_file(aait_store_remote, target, output)
    dialog_progress.stop()
def get_size(path):
    """
    Returns the size of the specified file or directory in bytes.

    :param path: Path to the file or directory.
    :return: Size in bytes.
    """
    if os.path.isfile(path):
        return os.path.getsize(path)
    elif os.path.isdir(path):
        return get_dir_size(path)
    else:
        SimpleDialogQt.BoxError("The specified path is neither a file nor a directory"+path)
        raise Exception("The specified path is neither a file nor a directory"+path)

def get_dir_size(directory):
    """
    Returns the total size of all files in a directory in bytes.

    :param directory: Path to the directory.
    :return: Size in bytes.
    """
    total_size = 0
    for dirpath, dirnames, filenames in os.walk(directory):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            # Ignore symbolic links
            if not os.path.islink(fp):
                total_size += os.path.getsize(fp)
    return total_size

def delete_path(path):
    """
    Deletes the specified file or directory.

    :param path: Path to the file or directory to be deleted.
    """
    try:
        if os.path.isfile(path):
            os.remove(path)  # Remove the file
            print(f"File '{path}' has been deleted.")
        elif os.path.isdir(path):
            shutil.rmtree(path)  # Remove the directory and all its contents
            print(f"Directory '{path}' and all its contents have been deleted.")
        else:
            print(f"The path '{path}' does not exist.")
    except Exception as e:
        SimpleDialogQt.BoxError("An error occurred while deleting "+path+"\""+str(e))
        raise("An error occurred while deleting "+path+"\""+str(e))

def already_downloaded(source,output):
    """
    Check if a file need to be downloaded or downloaded again
    based on file size
    remove file if necessary
    """
    # file or directory doesn't exist -> we need to download it

    if not os.path.exists(output):
        return False
    if get_size(source)==get_size(output):
        return True
    # remove output
    print("remove")
    delete_path(output)
    return False

def already_downloaded_compressed(repo_file,file_or_directory,output):
    """
    Check if a file need to be downloaded or downloaded again
    based on file size
    remove file if necessary
    """

    if not os.path.exists(output):
        return False
    output_size=get_size(output)
    input_size=0

    data = json.loads(aait_repo_file.decode(repo_file, "files_info.json"))
    if file_or_directory in data:
        input_size=int(data[file_or_directory])
    else:
        list_of_file = [cle for cle in data if cle.startswith(file_or_directory)]
        for element in list_of_file:
            input_size+=data[element]
    if input_size==output_size:
        return True
    print("remove")
    delete_path(output)
    return False


def already_downloaded_compressed_server(server_repo_url: str, file_or_directory: str, output: str) -> bool:

    """
    Checks whether the file or directory 'file_or_directory', listed on the server
    (described in the JSON file files_info.json), has already been downloaded locally to 'output',
    and if the local size matches the size declared in files_info.json.
    This is done using curl commands (via subprocess).

    :param server_repo_url: Base URL of the server directory,
                            e.g. "http://88.172.137.71:40386/share/ma0UJQ8-jqb9uqiM/"
    :param file_or_directory: Path (or key) representing the file or directory to check,
                              as referenced in the JSON.
    :param output: Local path to the file/directory where it was (or should be) downloaded.
    :return: True if the file/directory is already downloaded and up to date (same size),
             False otherwise.
    """

    # --- 1. Check if 'output' already exists locally
    if not Path(output).exists():
        #print(f"the path {output} doesn't exist locally")
        return False
    # --- 2. Get the local size
    output_size = get_size(output)

    # --- 3. Retrieve the expected size (input_size) from "files_info.json" on the server
    # Build the URL to the JSON file
    json_url = server_repo_url.rstrip('/') + '/files_info.json'

    # Download files_info.json using curl
    # -s : silent mode, -L : follow redirects
    process = subprocess.run(
        ["curl", "-s", "-L", json_url],
        capture_output=True
    )
    if process.returncode != 0:
        # If curl failed (bad URL, DNS error, etc.), display the error and exit
        err_msg = process.stderr.decode('utf-8', errors='replace')
        print(f"Failed to download {json_url}: {err_msg}")
        return False

    data_str = process.stdout.decode('utf-8', errors='replace')
    if not data_str.strip():
        # If the JSON file is empty or inaccessible
        print(f"The content of {json_url} is empty or inaccessible.")
        return False

    # Attempt to parse the JSON
    try:
        data = json.loads(data_str)
    except json.JSONDecodeError as e:
        print(f"The file {json_url} is not valid JSON: {e}")
        return False

    # --- 4. Compute input_size based on 'file_or_directory'
    input_size = 0
    if file_or_directory in data:
        # Exact key
        input_size = int(data[file_or_directory])
    else:
        # Assume file_or_directory is a directory: sum the sizes of all matching keys
        matching_keys = [k for k in data if k.startswith(file_or_directory)]
        for key in matching_keys:
            input_size += int(data[key])

    # --- 5. Compare
    if input_size == output_size:
        return True
    else:
        # If sizes do not match, remove the local folder to force a new download
        print(f"Local size ({output_size}) does not match the expected size ({input_size}).")
        print("remove")
        delete_path(output)
        return False


def exectute_python_post_install(script_path):
    """
    execute a one time scipt
    """
    try:
        local_aait_store_path = get_local_store_path()
    except Exception:
        SimpleDialogQt.BoxError("Impossible to open local ia store path")
        raise Exception("Impossible to open local ia store path")


    path_to_execute= local_aait_store_path +script_path
    if os.path.isfile(path_to_execute)==False:
        SimpleDialogQt.BoxError(path_to_execute+" not exist")
        raise Exception(path_to_execute+" not exist")

    sys.path.append(os.path.dirname(path_to_execute))
    le_module=__import__(Path(path_to_execute).stem)
    le_module.python_post_install()




def GetFromRemote(name_to_download, repo_if_necessary = None):
    """
    Get files form store ia
    return 0 if ok
    1 if not ok
    """
    if not repo_if_necessary:
        write_local_current_version()

        # Get all repositories from the file
        local_store_path = get_local_store_path()
        path_file = os.path.join(local_store_path, "remote_ressources_path.txt")
        repositories = []

        if os.path.exists(path_file):
            with open(path_file, 'r') as f:
                repositories = [line.strip() for line in f if line.strip()]
        if not repositories:
            SimpleDialogQt.BoxError("No repositories found!")
            return 1

        # Try each repository until we find the package
        for repo in repositories:
            remote_version = get_aait_store_remote_version(repo)
            local_version = get_aait_store_remote_version(get_local_store_path())

            if remote_version != local_version:
                SimpleDialogQt.BoxWarning("you need to update your ia store add on!")
                raise Exception("you need to update your ia store add on!")

            try:
                requirements = get_aait_store_requirements_json(repo_path=repo)

                ia_store_path = repo  # Use repo directly instead of getting it from the file
                delete_temporary_files = False
                downloaded = False

                for element in requirements:
                    if element['name'] != name_to_download:
                        continue


                    downloaded = True

                    for key in ("workflows", "extra_datas", "temporary_files"):
                        if key in element:
                            if key == "temporary_files":
                                delete_temporary_files = True
                            for subpath in element[key]:
                                if not subpath:
                                    continue
                                if not is_url(ia_store_path + "/" + subpath):
                                    download_aait_store_file_or_directory(server_path=ia_store_path, target=subpath, repo_path=repo)
                                else:
                                    download_aait_store_file_or_directory(server_path=ia_store_path, target=subpath, repo_path=repo)

                    if "python_post_install" in element:
                        for script in element["python_post_install"]:
                            if script:
                                exectute_python_post_install(script)

                    if downloaded:
                        if delete_temporary_files:
                            delete_path(get_local_store_path()+"temporary_files")
                        SimpleDialogQt.BoxInfo("Finished!")
                        return 0

            except Exception as e:
                print(e)
                continue
    else:
        write_local_current_version()

        # Get all repositories from the file
        local_store_path = get_local_store_path()
        path_file = os.path.join(local_store_path, "remote_ressources_path.txt")
        repositories = []

        if os.path.exists(path_file):
            with open(path_file, 'r') as f:
                repositories = [line.strip() for line in f if line.strip()]
        if not repositories:
            SimpleDialogQt.BoxError("No repositories found!")
            return 1

        # Try each repository until we find the package
        for repo in repositories:
            remote_version = get_aait_store_remote_version(repo)
            local_version = get_aait_store_remote_version(get_local_store_path())


            if remote_version != local_version:
                SimpleDialogQt.BoxWarning("you need to update your ia store add on!")
                raise Exception("you need to update your ia store add on!")

            try:
                requirements = get_aait_store_requirements_json(repo_path=repo)

                ia_store_path = repo  # Use repo directly instead of getting it from the file
                delete_temporary_files = False
                downloaded = False
                for element in requirements:
                    if element['name'] != name_to_download:
                        continue

                    downloaded = True

                    for key in ("workflows", "extra_datas", "temporary_files"):
                        if key in element:
                            if key == "temporary_files":
                                delete_temporary_files = True
                            for subpath in element[key]:
                                if not subpath:
                                    continue
                                if not is_url(ia_store_path + "/" + subpath):
                                    download_aait_store_file_or_directory(server_path=ia_store_path, target=subpath,
                                                                          repo_path=repo)
                                else:
                                    download_aait_store_file_or_directory(server_path=ia_store_path, target=subpath,
                                                                          repo_path=repo)

                    if "python_post_install" in element:
                        for script in element["python_post_install"]:
                            if script:
                                exectute_python_post_install(script)

                    if downloaded:
                        if delete_temporary_files:
                            delete_path(get_local_store_path() + "temporary_files")
                        SimpleDialogQt.BoxInfo("Finished!")
                        return 0

            except Exception as e:
                print(e)
                continue
            
    SimpleDialogQt.BoxError(f"Package {name_to_download} doesn't exist in any of your stores!")
    return 1

def TransfromPathToStorePath(path_in):
    """
    transform if possible an absolute path to a path relative in the store
    """
    path_in=path_in.replace("\\","/")
    local_store=get_local_store_path()
    if len(local_store)<2:
        return path_in
    if len(path_in)<len(local_store):
        return path_in
    if len(path_in)==len(local_store):
        return path_in
    if path_in[:len(local_store)]==local_store:
        return path_in[len(local_store):]
    return path_in


def TransfromStorePathToPath(path_in):
    """
    transform if possible a path relative in the store to an absolute path to
    """
    path_in=path_in.replace("\\","/")
    if len(path_in)<2:
        return get_local_store_path()+path_in
    if path_in[1]==":":# exemple c:
        return path_in
    if path_in[0]=="/": # exemple /dev or //serveur_name
        return path_in
    return get_local_store_path()+path_in

def getTempDir():
    """
    get temporary dir (available on nt, macosx and gnu/linux
    """
    temp_dir =os.getenv('TMPDIR') or os.getenv('TEMP') or  tempfile
    if temp_dir is None:
        raise "temp dir check not implementing in your OS"
    return str(temp_dir)


def reset_folder(folder_path, attempts=10, delay=0.05,recreate=True):
    """
    Attempts to delete a folder and its contents up to a specified number of times.
    If the folder is successfully deleted, it is then recreated.
    Returns 0 if successful, 1 if deletion fails after all attempts or if recreation fails.
    """
    for _ in range(attempts):
        if os.path.exists(folder_path):
            try:
                shutil.rmtree(folder_path)
                break  # Exit loop if deletion succeeds
            except Exception as e:
                print(f"Failed to delete folder: {e}")
                time.sleep(delay)

    if os.path.exists(folder_path):
        print(f"Failed to delete folder after {attempts} attempts.")
        return 1  # Failure if the folder still exists
    if recreate==False:
        return 0
    try:
        os.makedirs(folder_path, exist_ok=True)
        return 0  # Success
    except Exception as e:
        print(f"Failed to create folder: {e}")
        return 1  # Failure

def reset_files(files, attempts=10, delay=0.05):
    # files [file1,file2,file3]???
    for file in files:
        for _ in range(attempts):
            if os.path.exists(file):
                try:
                    os.remove(file)
                    break  # Exit loop if deletion succeeds
                except Exception as e:
                    print(f"Failed to delete folder: {e}")
                    time.sleep(delay)
        if os.path.exists(file):
            print(f"Failed to delete {file} after {attempts} attempts.")
            return 1  # Failure if the folder still exists
    return 0

def get_api_local_folder(workflow_id=""):
    aait_path = get_local_store_path()
    result=aait_path + "exchangeApi/"
    if workflow_id=="":
        return result
    if len(workflow_id)>4:
        if workflow_id[-4:]==".ows":
            workflow_id=workflow_id[:-4]
    result+=workflow_id+"/"
    return result

def get_api_local_folder_admin():
    return get_api_local_folder()[:-1]+"_adm/"

def get_api_local_folder_admin_locker():
    return get_api_local_folder()[:-1]+"_adm/locker/"

def get_path_linkHTMLWorkflow():
    aait_path = get_local_store_path()
    return aait_path + "Parameters/linkHTMLWorkflow/"

def get_path_mailFolder():
    aait_path = get_local_store_path()
    the_path= aait_path + "exchangeMail/"
    if not os.path.exists(the_path):
        os.makedirs(the_path)
    return the_path

def get_secret_content_dir():
    aait_path = get_local_store_path()
    the_path=aait_path+"/keys/"
    if not os.path.exists(the_path):
        os.makedirs(the_path)
    return the_path

def get_second_from_1970():
    return int(time.time())

def write_file_time(path):
    time= get_second_from_1970()
    time_ok=path[:-4]+".ok"
    reset_files([time_ok])
    with open(path, "w") as f:
        f.write(str(time))
        f.flush()
    try:
        with open(time_ok, "w"):
            pass
    except Exception as e:
        print(f"Error creating .ok file: {e}")
    return
def write_file_arbitrary_time(path,time):
    time= int(time)
    time_ok=path[:-4]+".ok"
    reset_files([time_ok])
    with open(path, "w") as f:
        f.write(str(time))
        f.flush()
    try:
        with open(time_ok, "w"):
            pass
    except Exception as e:
        print(f"Error creating .ok file: {e}")
    return

def read_file_time(path):
    time_ok = path[:-4] + ".ok"
    for _ in range(100):
        if not os.path.exists(time_ok):
                time.sleep(0.5)
    if  not os.path.exists(time_ok):
        return 0 # time out
    with open(path, "r") as f:
        content = f.read()
    if content.isdigit()==False:
        return int(get_second_from_1970()) # en cas d'erreur on renvoie le temps actuel
    return int(content)

def generate_unique_id_from_mac_timestamp():
    # Récupère l'adresse MAC de la machine (ou un fallback stable si MAC inaccessible)
    mac = uuid.getnode()
    # Temps Unix à la centieme de seconde (précision 0.01s)
    timestamp = int(time.time() * 100)
    # Conversion en octets
    mac_bytes = mac.to_bytes(6, byteorder='big')  # 6 octets = 48 bits
    ts_bytes = timestamp.to_bytes(8, byteorder='big')  # 6 octet je suis large
    # Concaténation binaire + encodage en base64 URL-safe sans padding
    raw_bytes = mac_bytes + ts_bytes
    out= str(base64.b32encode(raw_bytes).decode().rstrip('=').lower())
    return out

def describe_orange_table(data_table):
    """
    Décrit une Orange.data.Table.
    Retour:
      - None en cas d'erreur (mauvais type, exception, etc.)
      - (n_rows, columns) sinon, où:
          n_rows: int, nombre de lignes
          columns: liste de dicts {name, kind, var_type}
                   - kind ∈ {"feature","target","meta"}
                   - var_type ∈ {"StringVariable","ContinuousVariable","Categorical","TimeVariable", ...}
    """
    try:
        if not isinstance(data_table, Table):
            return None

        def var_type_name(var):
            return type(var).__name__         # fallback propre

        n_rows = len(data_table)
        dom = data_table.domain

        cols = []
        # Features
        for v in dom.attributes:
            cols.append({"name": v.name, "kind": "feature", "var_type": var_type_name(v)})
        # Targets (peut être 0, 1 ou plusieurs class_vars)
        for v in dom.class_vars:
            cols.append({"name": v.name, "kind": "target", "var_type": var_type_name(v)})
        # Metas
        for v in dom.metas:
            cols.append({"name": v.name, "kind": "meta", "var_type": var_type_name(v)})

        return n_rows, cols

    except Exception as e:
        print(e)
        return None


def create_trigger_table():
    var = StringVariable("Trigger")
    dom = Domain([], metas=[var])
    table = Table.from_list(domain=dom, rows=[["Trigger"]])
    return table


if __name__ == "__main__":
    # avant faire un bouton
    # set_aait_store_remote_ressources_path(ressource_path)
    # get_aait_store_remote_version("http://88.172.137.71:40386/share/ma0UJQ8-jqb9uqiM/AAIT_v240916")
    #
    # #download_aait_store_file_or_directory("http://88.172.137.71:40386/share/ma0UJQ8-jqb9uqiM/AAIT_v240916")
    # aait_store_remote = get_aait_store_remote_ressources_path()
    # print(aait_store_remote)
    # aait_store_local = get_local_store_path()
    # get_aait_store_requirements_json()
    # aait_store_content = get_aait_store_requirements_json()
    #print(already_downloaded_compressed_server("http://88.172.137.71:40386/share/ma0UJQ8-jqb9uqiM/AAIT_v240916","Models/NLP/all-mpnet-base-v2/config.json",r"C:\Users\timot\aait_store\Models\NLP\all-mpnet-base-v2\config.json"))
    #GetFromRemote("All AI Store")
    #IsStoreCompressed("http://88.172.137.71:40386/share/ma0UJQ8-jqb9uqiM/IFIA_models/repository.aait")
    #set_aait_store_remote_ressources_path("http://88.172.137.71:40386/share/ma0UJQ8-jqb9uqiM/IFIA_models/repository.aait")
    aaa = get_second_from_1970()
    print(aaa)




from Orange.data import Table, Domain, Instance


def unlink_domain(table: Table) -> Table:
    dom = table.domain

    # Recrée les variables sans compute_value
    new_attrs  = [v.copy(compute_value=None) for v in dom.attributes]
    new_cls    = [v.copy(compute_value=None) for v in dom.class_vars]
    new_metas  = [v.copy(compute_value=None) for v in dom.metas]
    new_domain = Domain(new_attrs, new_cls, new_metas)

    instances = []
    for row in table:
        # on évalue toutes les valeurs ici (les compute_value sont calculés)
        attr_vals = [row[v] for v in dom.attributes]
        cls_vals  = [row[v] for v in dom.class_vars]
        inst = Instance(new_domain, attr_vals + cls_vals)

        # on renseigne les métas via l’indexation
        for old_m, new_m in zip(dom.metas, new_metas):
            inst[new_m] = row[old_m]

        instances.append(inst)

    return Table.from_list(new_domain, instances)


import os
import sys
import time
import re

# from AnyQt import QtWidgets
from AnyQt.QtCore import QTimer,Qt
from AnyQt.QtWidgets import QApplication  # QMainWindow, QFileDialog
from AnyQt.QtWidgets import (QComboBox, QDialog, QGroupBox, QHBoxLayout,
                             QLabel, QPushButton,  QVBoxLayout, QListWidget, QLineEdit)
from Orange.widgets import widget

if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.fix_torch import fix_torch_dll_error
    from Orange.widgets.orangecontrib.AAIT.utils import (MetManagement,
                                                         SimpleDialogQt)
    from Orange.widgets.orangecontrib.AAIT.utils.MetManagement import GetFromRemote
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
else:
    from orangecontrib.AAIT.fix_torch import fix_torch_dll_error
    from orangecontrib.AAIT.utils import (MetManagement,
                                         SimpleDialogQt)
    from orangecontrib.AAIT.utils.MetManagement import GetFromRemote
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
fix_torch_dll_error.fix_error_torch()

class RepositoryManager(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Repository Manager")
        self.setModal(True)
        self.setMinimumWidth(500)  # Set minimum width for better readability
        
        # Main layout
        layout = QVBoxLayout()
        self.setLayout(layout)
        
        # Repository list
        list_label = QLabel("Current Repositories:")
        layout.addWidget(list_label)
        self.repo_list = QListWidget()
        layout.addWidget(self.repo_list)
        
        # Add Repository Group Box
        add_group = QGroupBox("Add Repository")
        add_layout = QVBoxLayout()
        add_group.setLayout(add_layout)
        
        # File selection
        file_layout = QHBoxLayout()
        self.file_button = QPushButton("Select New Repository")
        self.file_button.clicked.connect(self.add_repository)
        file_layout.addWidget(self.file_button)
        add_layout.addLayout(file_layout)
        
        # URL input
        url_layout = QHBoxLayout()
        url_label = QLabel("Or enter repository URL:")
        self.url_input = QLineEdit()
        self.url_input.setPlaceholderText("Enter URL here...")
        self.url_add_button = QPushButton("Add URL")
        self.url_add_button.clicked.connect(self.add_url_repository)
        
        url_layout.addWidget(url_label)
        url_layout.addWidget(self.url_input)
        url_layout.addWidget(self.url_add_button)
        add_layout.addLayout(url_layout)
        
        layout.addWidget(add_group)
        
        # Remove button
        self.remove_button = QPushButton("Remove Selected Repository")
        self.remove_button.clicked.connect(self.remove_repository)
        layout.addWidget(self.remove_button)
        
        # Load existing repositories
        self.load_repositories()
        
    def load_repositories(self):
        """Load existing repositories into the list"""
        self.repo_list.clear()
        repositories = self.parent().current_repositories
        self.repo_list.addItems(repositories)
    
    def add_repository(self):
        """Add a new repository"""
        folder = MetManagement.get_aait_store_remote_ressources_path()
        file = SimpleDialogQt.BoxSelectExistingFile(self, default_dir=folder, extention="Aiit file (*.aait)")
        if file == "":
            return
            
        if MetManagement.get_size(file) == 0:
            folder = os.path.dirname(os.path.abspath(file)).replace("\\", "/")
            if folder == "":
                return
            if folder[-1] != "/":
                folder += "/"
            if folder not in self.parent().current_repositories:
                self.parent().current_repositories.append(folder)
                self.repo_list.addItem(folder)
                self.save_repositories_to_file()
        else:
            # compressed case
            file = file.replace("\\", "/")
            if file not in self.parent().current_repositories:
                self.parent().current_repositories.append(file)
                self.repo_list.addItem(file)
                self.save_repositories_to_file()
        
        self.parent().update_requirements()
    
    def remove_repository(self):
        """Remove selected repository"""
        current_item = self.repo_list.currentItem()
        if current_item:
            repo = current_item.text()
            self.parent().current_repositories.remove(repo)
            self.repo_list.takeItem(self.repo_list.row(current_item))
            self.save_repositories_to_file()
            self.parent().update_requirements()

    def add_url_repository(self):
        """Add a new repository from URL"""
        url = self.url_input.text().strip()
        if url.endswith('.aait') and not MetManagement.IsStoreCompressed(url):
            temp_url = url.split("/")[:-1]
            url = "/".join(temp_url) + "/"
        if not url:
            SimpleDialogQt.BoxWarning("Please enter a URL")
            return
            
        # Basic URL validation
        if not url.startswith(('http://', 'https://')):
            SimpleDialogQt.BoxWarning("Please enter a valid URL starting with http:// or https://")
            return
            
        # Use the same logic as file paths
        if url not in self.parent().current_repositories:
            self.parent().current_repositories.append(url)
            self.repo_list.addItem(url)
            self.url_input.clear()  # Clear the input field
            self.save_repositories_to_file()
            self.parent().update_requirements()

    def save_repositories_to_file(self):
        """Save current repositories to remote_ressources_path.txt"""
        try:
            local_store_path = MetManagement.get_local_store_path()
            path_file = os.path.join(local_store_path, "remote_ressources_path.txt")

            # Create a lock file
            lock_file = path_file + ".lock"
            while os.path.exists(lock_file):
                time.sleep(0.1)
            
            # Create lock
            with open(lock_file, 'w') as f:
                f.write('locked')
            
            try:
                # Write all repositories to the file
                with open(path_file, 'w') as f:
                    for repo in self.parent().current_repositories:
                        f.write(repo + '\n')
                
            finally:
                # Remove lock
                if os.path.exists(lock_file):
                    os.remove(lock_file)
        except Exception as e:
            print(f"Error saving repositories: {e}")

@apply_modification_from_python_file(filepath_original_widget=__file__)
class OWAAITResourcesManager(widget.OWWidget):
    name = "AAIT Resources Manager"
    description = "Manage AAIT resources, such as model, example workflows, datasets...."
    icon = "icons/documents.png"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/documents.png"
    priority = 1001
    # Path
    dossier_du_script = os.path.dirname(os.path.abspath(__file__))

    def __init__(self):
        super().__init__()
        self.requirements = []  # Changed to list to store multiple repositories' requirements
        self.current_repositories = []  # Store paths of selected repositories
        self.controlAreaVisible = False

    # trigger if standard windows is opened
    def showEvent(self, event):
        super().showEvent(event)
        self.show_dialog()
        # We cannot close the standard ui widget it is displayed
        # so it makes a little tinkles :(
        QTimer.singleShot(0, self.close)
    def open_explorer(self):
        path = MetManagement.get_local_store_path()
        if sys.platform.startswith("Darwin") or sys.platform.startswith("darwin"):
            import subprocess
            subprocess.run(["open", path])
            return
        elif sys.platform.startswith("win"):
            os.startfile(path)
            return
        # other platfoprm -> nothing

    def show_dialog(self):
        # third-party code execution vs standard code execution
        if False == os.path.isfile(MetManagement.get_local_store_path() + "AddOn/prefix_show_dialog.py"):
            dialog = QDialog()
            layout_a = QVBoxLayout()
            dialog.setLayout(layout_a)
            model = None
        else:
            sys.path.append(MetManagement.get_local_store_path() + "AddOn")
            import prefix_show_dialog
            stable_dependency = True
            if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
                stable_dependency = False
            dialog, model = prefix_show_dialog.prefix_dialog_function(self,stable_dependency)

        # download section
        main_layout = QVBoxLayout()
        group_box = QGroupBox("Download new minimum working example")
        group_layout = QVBoxLayout()

        # Elements are presented horizontally
        h_layout = QHBoxLayout()
        v_layout_button_combo_box = QVBoxLayout()
        
        # Add repository management button
        self.manage_repos_button = QPushButton('Manage Repositories')
        self.manage_repos_button.clicked.connect(self.show_repository_manager)
        v_layout_button_combo_box.addWidget(self.manage_repos_button)
        
        # Add combo box and other elements
        self.comboBox = QComboBox()
        self.comboBox.setMinimumSize(200, 10)
        v_layout_button_combo_box.addWidget(self.comboBox)
        
        self.saveButton = QPushButton('Download')
        self.label_info = QLabel('')
        v_layout_button_combo_box.addWidget(self.saveButton)
        v_layout_button_combo_box.addWidget(self.label_info)

        # Add the vertical layout to horizontal layout
        h_layout.addLayout(v_layout_button_combo_box)
        
        # Add h_layout to group_layout
        group_layout.addLayout(h_layout)
        group_box.setLayout(group_layout)
        main_layout.addWidget(group_box)
        # self.selectableLabel = QLabel(f"local store {MetManagement.get_local_store_path() }")
        # self.selectableLabel.setTextInteractionFlags(Qt.TextSelectableByMouse)
        # main_layout.addWidget(self.selectableLabel)


        # Nouveau layout horizontal
        h_layout2 = QHBoxLayout()

        # Chemin du dossier local
        local_path = MetManagement.get_local_store_path()

        # Bouton "local store"
        self.open_button = QPushButton("open local store")
        self.open_button.clicked.connect(self.open_explorer)
        h_layout2.addWidget(self.open_button)

        # Label sélectionnable
        self.selectableLabel = QLabel(local_path)
        self.selectableLabel.setTextInteractionFlags(Qt.TextSelectableByMouse)
        h_layout2.addWidget(self.selectableLabel)

        # Ajouter au layout principal
        main_layout.addLayout(h_layout2)



        main_layout.setContentsMargins(5, 5, 5, 5)
        dialog.layout().insertLayout(0, main_layout)

        # Connect signals
        self.comboBox.currentIndexChanged.connect(self.handleComboBoxChange)
        self.saveButton.clicked.connect(self.saveFile)
        
        # Initialize repositories
        self.load_repositories()

        if False == os.path.isfile(MetManagement.get_local_store_path() + "AddOn/postfix_show_dialog.py"):
            dialog.exec()
        else:
            sys.path.append(MetManagement.get_local_store_path() + "AddOn")
            import postfix_show_dialog
            stable_dependency = True
            if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
                stable_dependency = False
            postfix_show_dialog.postfix_dialog_function(dialog, model)

    def load_repositories(self):
        """Load repositories from remote_ressources_path.txt"""
        try:
            local_store_path = MetManagement.get_local_store_path()
            path_file = os.path.join(local_store_path, "remote_ressources_path.txt")
            
            if os.path.exists(path_file):
                with open(path_file, 'r') as f:
                    repositories = [line.strip() for line in f if line.strip()]
                    self.current_repositories = repositories
            else:
                # If file doesn't exist, start with default repository
                default_repo = MetManagement.get_aait_store_remote_ressources_path()
                if default_repo:
                    self.current_repositories = [default_repo]
                    # Create the file with default repository
                    with open(path_file, 'w') as f:
                        f.write(default_repo + '\n')
                else:
                    self.current_repositories = []
        except Exception as e:
            print(f"Error loading repositories: {e}")
            self.current_repositories = []
        
        self.update_requirements()
        self.populate_combo_box()
        

    def populate_combo_box(self):
        # clean combo box if we change of repository
        self.comboBox.clear()
        workflows = []
        descriptions = dict()
        

        if not self.requirements:
            return
            
        for element in self.requirements:
            # Get repository name - handle both file and folder paths
            repo_path = element['repository']
            if repo_path.endswith('.aait'):
                repo_name = os.path.basename(repo_path)
            else:
                # For folder paths, get the last directory name
                repo_name = os.path.basename(os.path.dirname(repo_path))
            
            name = f"{element['name']} ({repo_name})"
            workflows.append(name)
            descriptions[name] = element["description"][0]

        self.descriptions = descriptions
        print(self.descriptions)
        self.comboBox.addItems(workflows)

        if workflows:
            self.handleComboBoxChange(0)

    def handleComboBoxChange(self, index):
        selected_file = self.comboBox.itemText(index)
        if selected_file == "":
            self.label_info.setText("")
            return
        # Set the label to the description
        description = self.descriptions.get(selected_file, "")
        self.label_info.setText(description)

    def read_description(self, file_name):
        # Chemin du fichier texte contenant la description
        description_file_path = os.path.join(self.dossier_du_script, 'ows_example',
                                             f'{os.path.splitext(file_name)[0]}.txt')
        # Lire le contenu du fichier s'il existe, sinon retourner une chaîne vide
        if os.path.exists(description_file_path):
            with open(description_file_path, 'r') as file:
                description = file.read()
        else:
            description = ""
        print("alllloooo: ", description)
        return description

    def saveFile(self):
        # Get selected file and remove repository info from display name
        selected_display = self.comboBox.currentText()
        match = re.search(r'\((.*?)\)', selected_display)
        reposit = match.group(1) if match else ''
        if not reposit.endswith('.aait'):
            selected_display = re.sub(r'\(.*?\)', '()', selected_display)
        current_repos = self.current_repositories.copy()

        # Find the corresponding requirement
        selected_requirement = None
        for req in self.requirements:
            display_name = f"{req['name']} ({os.path.basename(req['repository'])})"
            if MetManagement.IsStoreCompressed(req['repository']):
                repository = req['repository'].split("/")[-1]
            else:
                repository = req['repository'].split("/")[-2]
            if display_name == selected_display and repository == reposit:
                selected_requirement = req
                break
        
        if selected_requirement:
            self.label_info.setText('Synchronization in progress')
            QApplication.processEvents()  # Update UI
            # Set the correct repository for download
            MetManagement.set_aait_store_remote_ressources_path(selected_requirement['repository'])
            
            # Get just the name without repository info
            file_name = selected_requirement['name']

            try:
                # Call GetFromRemote with the local store path as target
                # target_path = MetManagement.get_local_store_path()
                GetFromRemote(file_name, repo_if_necessary=reposit)
                self.label_info.setText('Download completed')
            except Exception as e:
                self.label_info.setText(f'Error: {str(e)}')
            finally:
                QApplication.processEvents()  # Update UI
        try:
            local_store_path = MetManagement.get_local_store_path()
            path_file = os.path.join(local_store_path, "remote_ressources_path.txt")
            
            # Create a lock file
            lock_file = path_file + ".lock"
            while os.path.exists(lock_file):
                time.sleep(0.1)
            
            # Create lock
            with open(lock_file, 'w') as f:
                f.write('locked')
            
            try:
                # Write all repositories to the file
                with open(path_file, 'w') as f:
                    for repo in current_repos:
                        f.write(repo + '\n')
            finally:
                # Remove lock
                if os.path.exists(lock_file):
                    os.remove(lock_file)
                    
        except Exception as e:
            print(f"Error saving repositories: {e}")
            

    def show_repository_manager(self):
        """Show the repository manager dialog"""
        dialog = RepositoryManager(self)
        dialog.exec()

    def update_requirements(self):
        """Update requirements from all repositories"""
        self.requirements = []
        # Store the current repositories before we start

        current_repos = self.current_repositories.copy()

        for repo in current_repos:
            # Set the remote resources path temporarily
            MetManagement.set_aait_store_remote_ressources_path(repo)
            repo_requirements = MetManagement.get_aait_store_requirements_json()
            if repo_requirements:
                # Add repository information to each requirement
                for req in repo_requirements:
                    req['repository'] = repo
                self.requirements.extend(repo_requirements)
        
        # Write repositories back to file
        try:
            local_store_path = MetManagement.get_local_store_path()
            path_file = os.path.join(local_store_path, "remote_ressources_path.txt")
            
            # Create a lock file
            lock_file = path_file + ".lock"
            while os.path.exists(lock_file):
                time.sleep(0.1)
            
            # Create lock
            with open(lock_file, 'w') as f:
                f.write('locked')
            
            try:
                # Write all repositories to the file
                with open(path_file, 'w') as f:
                    for repo in current_repos:
                        f.write(repo + '\n')
            finally:
                # Remove lock
                if os.path.exists(lock_file):
                    os.remove(lock_file)
                    
        except Exception as e:
            print(f"Error saving repositories: {e}")
            
        self.populate_combo_box()

if __name__ == '__main__':
    app = QApplication(sys.argv)
    window = OWAAITResourcesManager()
    window.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()




import os
from pathlib import Path
from typing import List
import shutil
import time
if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils import MetManagement
else:
    from orangecontrib.AAIT.utils import MetManagement



def DeleteFolderlistOrFilelist(paths: list[str]):
    """
    supprime des fichier et des dossier dans la liste
    """
    for p in paths:
        if os.path.exists(p):
            if os.path.isfile(p):
                MetManagement.reset_files([p], attempts=10, delay=0.05)
            elif os.path.isdir(p):
                MetManagement.reset_folder(p, attempts=10, delay=0.05, recreate=False)

def ensure_dirs(paths: List[str]):
    """
    Crée récursivement tous les dossiers donnés dans la liste.
    Ne lève pas d'erreur si le dossier existe déjà.
    Compatible Windows, macOS, Linux.
    """
    for p in paths:
        path = Path(p).expanduser()
        if path.exists() and not path.is_dir():
            continue  # un fichier existe déjà ici, on ne touche pas
        os.makedirs(path, exist_ok=True)

def move_or_rename(src_list: List[str], dst_list: List[str]):
    """
    Déplace ou renomme des fichiers / dossiers.
    - src_list et dst_list doivent avoir la même longueur.
    - Crée les dossiers parents de destination si besoin.
    - Ne produit aucune sortie (fonction silencieuse).
    - Compatible Windows, macOS, Linux.
    """
    if len(src_list) != len(dst_list):
        raise ValueError("Les listes source et destination doivent avoir la même longueur.")

    for src, dst in zip(src_list, dst_list):
        src_path = Path(src).expanduser()
        dst_path = Path(dst).expanduser()

        if not src_path.exists():
            continue  # ignore si le fichier/dossier source n'existe pas

        dst_path.parent.mkdir(parents=True, exist_ok=True)  # crée le dossier parent si besoin
        os.replace(src_path, dst_path)  # déplace ou renomme (remplace si existe)

def copy_and_overwrite(src_list: List[str], dst_list: List[str]):
    """
    Copie des fichiers ou dossiers (récursivement pour les dossiers).
    - src_list et dst_list doivent avoir la même longueur.
    - Crée automatiquement les dossiers parents de destination.
    - Écrase les fichiers existants sans erreur.
    - Fonction silencieuse (aucune sortie console).
    - Compatible Windows, macOS, Linux.
    """
    if len(src_list) != len(dst_list):
        raise ValueError("Les listes source et destination doivent avoir la même longueur.")

    for src, dst in zip(src_list, dst_list):
        src_path = Path(src).expanduser()
        dst_path = Path(dst).expanduser()

        if not src_path.exists():
            continue  # ignore les sources inexistantes

        dst_path.parent.mkdir(parents=True, exist_ok=True)

        if src_path.is_dir():
            # Copie récursive du dossier (remplace si déjà présent)
            if dst_path.exists():
                shutil.rmtree(dst_path)
            shutil.copytree(src_path, dst_path)
        else:
            # Copie simple de fichier
            shutil.copy2(src_path, dst_path)

def sleep_seconds(seconds: float):
    """
    Suspend l'exécution pendant 'seconds' secondes.
    Compatible Windows, macOS, Linux.
    """
    if seconds <= 0:
        return
    time.sleep(seconds)


import os
from pathlib import Path
import sys
import platform

import Orange.data
from Orange.data import Table, Domain, StringVariable, ContinuousVariable
from AnyQt.QtWidgets import QApplication
from Orange.widgets import widget
from Orange.widgets.utils.signals import Input, Output

if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
    from Orange.widgets.orangecontrib.AAIT.utils import thread_management
else:
    from orangecontrib.AAIT.utils.import_uic import uic
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
    from orangecontrib.AAIT.utils import thread_management


@apply_modification_from_python_file(filepath_original_widget=__file__)
class OWFileSyncChecker(widget.OWWidget):
    name = "File Sync Checker"
    description = 'Verify if the files contained in Data are the same as the files contained in Reference. The verification is done thanks to both the "path" and "file size" columns.'
    category = "AAIT - TOOLBOX"
    icon = "icons/owfilesyncchecker.svg"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/owfilesyncchecker.svg"
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owfilesyncchecker.ui")
    want_control_area = False
    priority = 1060

    class Inputs:
        data = Input("Data", Orange.data.Table)
        reference = Input("Reference", Orange.data.Table)

    class Outputs:
        data = Output("Files only in Data", Orange.data.Table)
        processed = Output("Files in Data & Reference", Orange.data.Table)

    @Inputs.data
    def set_data(self, in_data):
        self.data = in_data
        if self.autorun:
            self.run()

    @Inputs.reference
    def set_reference(self, in_reference):
        self.reference = in_reference
        if self.autorun:
            self.run()


    def __init__(self):
        super().__init__()
        # Qt Management
        self.setFixedWidth(470)
        self.setFixedHeight(300)
        uic.loadUi(self.gui, self)

        # Data Management
        self.data = None
        self.reference = "default"
        self.autorun = True
        self.thread = None
        self.result = None
        self.post_initialized()

    def run(self):
        self.warning("")
        self.error("")

        # If Thread is already running, interrupt it
        if self.thread is not None:
            if self.thread.isRunning():
                self.thread.safe_quit()

        if self.data is None:
            self.Outputs.data.send(None)
            self.Outputs.processed.send(None)
            return

        if self.reference == "default":
            self.Outputs.data.send(None)
            self.Outputs.processed.send(None)
            return

        if self.reference is None:
            self.warning('There is no Reference table. All the files in Data will be considered as new files.')
            self.Outputs.data.send(self.data)
            self.Outputs.processed.send(None)
            return

        if "path" not in self.data.domain or "path" not in self.reference.domain:
            self.error('You need a "path" column in both Data and Reference tables.')
            return

        if "file size" not in self.data.domain:
            self.warning('There is no "file size" column in your Data table. All the files in Data will be considered as new files.')
            self.Outputs.data.send(self.data)
            self.Outputs.processed.send(None)
            return

        if "file size" not in self.reference.domain:
            self.warning('There is no "file size" column in your Reference table. All the files in Data will be considered as new files.')
            self.Outputs.data.send(self.data)
            self.Outputs.processed.send(None)
            return

        # Start progress bar
        self.progressBarInit()

        # Start threading
        self.thread = thread_management.Thread(self.check_for_sync, self.data, self.reference)
        self.thread.progress.connect(self.handle_progress)
        self.thread.result.connect(self.handle_result)
        self.thread.finish.connect(self.handle_finish)
        self.thread.start()

    def check_for_sync(self, table_data, table_reference, progress_callback=None, argself=None):

        def fast_norm(path_str):
            p = Path(path_str).expanduser()
            try:
                p = p.absolute()  # no disk access
            except OSError:
                pass

            s = str(p)
            if os.name == "nt":
                return s.replace("/", "\\").lower()
            return s

        reference = table_reference.copy()

        # ---- Extract & normalize paths ----
        paths_data = {
            fast_norm(row["path"].value): int(row["file size"].value)
            for row in table_data
        }

        paths_ref = {}
        ref_row_for_path = {}
        for i, row in enumerate(reference):
            p = fast_norm(row["path"].value)
            paths_ref[p] = int(row["file size"].value)
            ref_row_for_path[p] = i

        progress_update(20, progress_callback)

        # ---- Compute aligned roots once ----
        def common_root(keys):
            if not keys:
                return Path(".")
            return get_common_path([Path(k) for k in keys])

        common_data = common_root(paths_data.keys())
        common_ref = common_root(paths_ref.keys())

        # ---- Align reference paths quickly ----
        aligned_ref = {}
        for p, size in paths_ref.items():
            try:
                rel = Path(p).relative_to(common_ref)
                new = str((common_data / rel).absolute())
            except Exception:
                new = p
            aligned_ref[new] = size


        progress_update(50, progress_callback)

        set_data = set(paths_data.keys())
        set_ref = set(aligned_ref.keys())

        common = set_data & set_ref
        only_data = set_data - set_ref
        only_ref = set_ref - set_data

        print(f"Files in common: {len(common)}")
        print(f"Only in data: {len(only_data)}")
        print(f"Only in ref: {len(only_ref)}")

        # ---- Collect process list + rows to remove ----
        files_to_process = []
        remove_indices = set()

        for p in common:
            if paths_data[p] != aligned_ref[p]:
                files_to_process.append([Path(p), paths_data[p]])
                if p in ref_row_for_path:
                    remove_indices.add(ref_row_for_path[p])

        progress_update(60, progress_callback)

        for p in only_data:
            files_to_process.append([Path(p), paths_data[p]])


        progress_update(70, progress_callback)

        for p in only_ref:
            if p in ref_row_for_path:
                remove_indices.add(ref_row_for_path[p])


        progress_update(80, progress_callback)

        # ---- Build new reference table once ----
        if remove_indices:
            keep_rows = [i for i in range(len(reference)) if i not in remove_indices]
            reference = reference[keep_rows]

        # ---- Re-align paths in reference ----
        for row in reference:
            p = fast_norm(row["path"].value)
            try:
                rel = Path(p).relative_to(common_ref)
                row["path"] = str((common_data / rel).absolute())
            except Exception:
                pass

        progress_update(100, progress_callback)

        # ---- Build output table ----
        path_var = StringVariable("path")
        size_var = ContinuousVariable("file size")
        dom = Domain([], metas=[path_var, size_var])
        out_data = Table.from_list(dom, files_to_process)

        return out_data, reference



    def handle_progress(self, value: float) -> None:
        self.progressBarSet(value)

    def handle_result(self, result):
        data = result[0]
        processed_data = result[1]
        try:
            self.Outputs.data.send(data)
            self.Outputs.processed.send(processed_data)
            self.data = None
            self.reference = "default"
        except Exception as e:
            print("An error occurred when sending out_data:", e)
            self.Outputs.data.send(None)
            return

    def handle_finish(self):
        self.progressBarFinished()


    def post_initialized(self):
        pass


def progress_update(value, progress_callback):
    if progress_callback is not None:
        progress_value = float(value)
        progress_callback(progress_value)


def remove_from_table(filepath, table):
    """
    Remove rows from the Orange table where 'path' matches the given filepath.
    """
    filepath = Path(filepath).resolve()

    filtered_table = Table.from_list(
        domain=table.domain,
        rows=[row for row in table
              if Path(str(row["path"].value)).resolve() != filepath]
    )
    return filtered_table


def get_common_path(paths):
    """
    Find the common root directory among a list of file paths.

    - If the list contains only one path, the parent directory of that path
      is returned (to ensure the result is always a directory).
    - If the list contains multiple paths, their deepest shared parent
      directory is returned using os.path.commonpath.

    Parameters
    ----------
    paths : list[pathlib.Path] or list[str]
        A list of file or directory paths.

    Returns
    -------
    pathlib.Path
        The common root directory as a Path object.
    """
    paths = [str(p) for p in paths]

    if len(paths) == 1:
        return Path(paths[0]).parent
    else:
        return Path(os.path.commonpath(paths))



if __name__ == "__main__":
    app = QApplication(sys.argv)
    my_widget = OWFileSyncChecker()
    my_widget.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()
